from __future__ import annotations

import datetime
from collections.abc import Mapping
from typing import Any, TypeVar, cast
from uuid import UUID

from attrs import define as _attrs_define
from dateutil.parser import isoparse

from ...client_types import UNSET, Unset

T = TypeVar("T", bound="SkuOptimizedResultsDto")


@_attrs_define
class SkuOptimizedResultsDto:
    """
    Attributes:
        id (int | Unset):
        sku_property_id (int | None | Unset):
        sku_id (int | None | Unset):
        location_id (int | None | Unset):
        channel_id (int | None | Unset):
        sku_grouping_id (int | None | Unset):
        sku_optimized_results_group_id (int | None | Unset):
        tenant_id (None | Unset | UUID):
        calculated_date_time (datetime.datetime | Unset):
        effective_to_date_time (datetime.datetime | None | Unset):
        reorder_point (float | None | Unset):
        order_quantity (float | None | Unset):
        lead_demand (float | None | Unset):
        forecast_period_demand (float | None | Unset):
        overridden_future_demand_effective_from_now (float | None | Unset):
        safety_stock_level (float | None | Unset):
        economic_order_quantity (float | None | Unset):
        optimial_stock_cycle (float | None | Unset):
        lead_time_days (int | None | Unset):
        reorder_frequency_days (int | None | Unset):
        order_count (int | Unset):
        latest_order_date (datetime.datetime | None | Unset):
        first_purchase_date (datetime.datetime | None | Unset):
        minimum_order_quantity (int | None | Unset):
        batch_size (int | None | Unset):
        stock_on_hand (float | None | Unset):
        stock_on_order (float | None | Unset):
        finished_good_stock_on_hand (float | None | Unset):
        finished_good_quantity_used (float | None | Unset):
        component_stock_on_hand (float | None | Unset):
        days_until_replenishment_due (int | None | Unset):
        days_until_stock_out (int | None | Unset):
        is_uncertain (bool | Unset):
        orders_in_previous_lead_time (float | Unset):
        avg_daily_orders_last_120_days (float | Unset):
        lead_demand_prediction_based_on_average (float | None | Unset):
        lead_demand_prediction_based_on_linear_regression (float | None | Unset):
        lead_demand_prediction_based_on_2_nd_order_polynomial_regression (float | None | Unset):
        lead_demand_prediction_based_on_last_month (float | None | Unset):
        lead_demand_prediction_based_on_previous_leadtime_days (float | None | Unset):
        standard_dev (float | None | Unset):
        service_factor (float | None | Unset):
        max_r_squared (float | None | Unset):
        min_r_squared (float | None | Unset):
        max_range (float | None | Unset):
        parent_quantity (float | None | Unset):
        most_accurate_algorithm_type_id (int | None | Unset):
        error_text (None | str | Unset):
        category (None | str | Unset):
        sub_category (None | str | Unset):
        brand (None | str | Unset):
        product_type (None | str | Unset):
        option1 (None | str | Unset):
        option2 (None | str | Unset):
        option3 (None | str | Unset):
        size (None | str | Unset):
        product_code (None | str | Unset):
        name (None | str | Unset):
        sku_cost (float | None | Unset):
        sku_price (float | None | Unset):
        is_discontinued (bool | Unset):
        is_unstocked (bool | Unset):
        product_of_count (int | Unset):
        component_of_count (int | Unset):
        customer_count (int | Unset):
        location_count (int | Unset):
        location_name (None | str | Unset):
        show_forecast_for_all_locations (bool | Unset):
        manufacturing_time (float | None | Unset):
        minimum_shelf_level (float | None | Unset):
        maximum_shelf_level (float | None | Unset):
        service_level (float | None | Unset):
        weight (float | None | Unset):
        height (float | None | Unset):
        width (float | None | Unset):
        length (float | None | Unset):
        dimensions_cubic_meters (float | None | Unset):
        external_id (None | str | Unset):
        sku_code (None | str | Unset):
        external_id_parent (None | str | Unset):
        sku_code_parent (None | str | Unset):
        child_variants_count (int | Unset):
    """

    id: int | Unset = UNSET
    sku_property_id: int | None | Unset = UNSET
    sku_id: int | None | Unset = UNSET
    location_id: int | None | Unset = UNSET
    channel_id: int | None | Unset = UNSET
    sku_grouping_id: int | None | Unset = UNSET
    sku_optimized_results_group_id: int | None | Unset = UNSET
    tenant_id: None | Unset | UUID = UNSET
    calculated_date_time: datetime.datetime | Unset = UNSET
    effective_to_date_time: datetime.datetime | None | Unset = UNSET
    reorder_point: float | None | Unset = UNSET
    order_quantity: float | None | Unset = UNSET
    lead_demand: float | None | Unset = UNSET
    forecast_period_demand: float | None | Unset = UNSET
    overridden_future_demand_effective_from_now: float | None | Unset = UNSET
    safety_stock_level: float | None | Unset = UNSET
    economic_order_quantity: float | None | Unset = UNSET
    optimial_stock_cycle: float | None | Unset = UNSET
    lead_time_days: int | None | Unset = UNSET
    reorder_frequency_days: int | None | Unset = UNSET
    order_count: int | Unset = UNSET
    latest_order_date: datetime.datetime | None | Unset = UNSET
    first_purchase_date: datetime.datetime | None | Unset = UNSET
    minimum_order_quantity: int | None | Unset = UNSET
    batch_size: int | None | Unset = UNSET
    stock_on_hand: float | None | Unset = UNSET
    stock_on_order: float | None | Unset = UNSET
    finished_good_stock_on_hand: float | None | Unset = UNSET
    finished_good_quantity_used: float | None | Unset = UNSET
    component_stock_on_hand: float | None | Unset = UNSET
    days_until_replenishment_due: int | None | Unset = UNSET
    days_until_stock_out: int | None | Unset = UNSET
    is_uncertain: bool | Unset = UNSET
    orders_in_previous_lead_time: float | Unset = UNSET
    avg_daily_orders_last_120_days: float | Unset = UNSET
    lead_demand_prediction_based_on_average: float | None | Unset = UNSET
    lead_demand_prediction_based_on_linear_regression: float | None | Unset = UNSET
    lead_demand_prediction_based_on_2_nd_order_polynomial_regression: (
        float | None | Unset
    ) = UNSET
    lead_demand_prediction_based_on_last_month: float | None | Unset = UNSET
    lead_demand_prediction_based_on_previous_leadtime_days: float | None | Unset = UNSET
    standard_dev: float | None | Unset = UNSET
    service_factor: float | None | Unset = UNSET
    max_r_squared: float | None | Unset = UNSET
    min_r_squared: float | None | Unset = UNSET
    max_range: float | None | Unset = UNSET
    parent_quantity: float | None | Unset = UNSET
    most_accurate_algorithm_type_id: int | None | Unset = UNSET
    error_text: None | str | Unset = UNSET
    category: None | str | Unset = UNSET
    sub_category: None | str | Unset = UNSET
    brand: None | str | Unset = UNSET
    product_type: None | str | Unset = UNSET
    option1: None | str | Unset = UNSET
    option2: None | str | Unset = UNSET
    option3: None | str | Unset = UNSET
    size: None | str | Unset = UNSET
    product_code: None | str | Unset = UNSET
    name: None | str | Unset = UNSET
    sku_cost: float | None | Unset = UNSET
    sku_price: float | None | Unset = UNSET
    is_discontinued: bool | Unset = UNSET
    is_unstocked: bool | Unset = UNSET
    product_of_count: int | Unset = UNSET
    component_of_count: int | Unset = UNSET
    customer_count: int | Unset = UNSET
    location_count: int | Unset = UNSET
    location_name: None | str | Unset = UNSET
    show_forecast_for_all_locations: bool | Unset = UNSET
    manufacturing_time: float | None | Unset = UNSET
    minimum_shelf_level: float | None | Unset = UNSET
    maximum_shelf_level: float | None | Unset = UNSET
    service_level: float | None | Unset = UNSET
    weight: float | None | Unset = UNSET
    height: float | None | Unset = UNSET
    width: float | None | Unset = UNSET
    length: float | None | Unset = UNSET
    dimensions_cubic_meters: float | None | Unset = UNSET
    external_id: None | str | Unset = UNSET
    sku_code: None | str | Unset = UNSET
    external_id_parent: None | str | Unset = UNSET
    sku_code_parent: None | str | Unset = UNSET
    child_variants_count: int | Unset = UNSET

    def to_dict(self) -> dict[str, Any]:
        id = self.id

        sku_property_id: int | None | Unset
        if isinstance(self.sku_property_id, Unset):
            sku_property_id = UNSET
        else:
            sku_property_id = self.sku_property_id

        sku_id: int | None | Unset
        if isinstance(self.sku_id, Unset):
            sku_id = UNSET
        else:
            sku_id = self.sku_id

        location_id: int | None | Unset
        if isinstance(self.location_id, Unset):
            location_id = UNSET
        else:
            location_id = self.location_id

        channel_id: int | None | Unset
        if isinstance(self.channel_id, Unset):
            channel_id = UNSET
        else:
            channel_id = self.channel_id

        sku_grouping_id: int | None | Unset
        if isinstance(self.sku_grouping_id, Unset):
            sku_grouping_id = UNSET
        else:
            sku_grouping_id = self.sku_grouping_id

        sku_optimized_results_group_id: int | None | Unset
        if isinstance(self.sku_optimized_results_group_id, Unset):
            sku_optimized_results_group_id = UNSET
        else:
            sku_optimized_results_group_id = self.sku_optimized_results_group_id

        tenant_id: None | str | Unset
        if isinstance(self.tenant_id, Unset):
            tenant_id = UNSET
        elif isinstance(self.tenant_id, UUID):
            tenant_id = str(self.tenant_id)
        else:
            tenant_id = self.tenant_id

        calculated_date_time: str | Unset = UNSET
        if not isinstance(self.calculated_date_time, Unset):
            calculated_date_time = self.calculated_date_time.isoformat()

        effective_to_date_time: None | str | Unset
        if isinstance(self.effective_to_date_time, Unset):
            effective_to_date_time = UNSET
        elif isinstance(self.effective_to_date_time, datetime.datetime):
            effective_to_date_time = self.effective_to_date_time.isoformat()
        else:
            effective_to_date_time = self.effective_to_date_time

        reorder_point: float | None | Unset
        if isinstance(self.reorder_point, Unset):
            reorder_point = UNSET
        else:
            reorder_point = self.reorder_point

        order_quantity: float | None | Unset
        if isinstance(self.order_quantity, Unset):
            order_quantity = UNSET
        else:
            order_quantity = self.order_quantity

        lead_demand: float | None | Unset
        if isinstance(self.lead_demand, Unset):
            lead_demand = UNSET
        else:
            lead_demand = self.lead_demand

        forecast_period_demand: float | None | Unset
        if isinstance(self.forecast_period_demand, Unset):
            forecast_period_demand = UNSET
        else:
            forecast_period_demand = self.forecast_period_demand

        overridden_future_demand_effective_from_now: float | None | Unset
        if isinstance(self.overridden_future_demand_effective_from_now, Unset):
            overridden_future_demand_effective_from_now = UNSET
        else:
            overridden_future_demand_effective_from_now = (
                self.overridden_future_demand_effective_from_now
            )

        safety_stock_level: float | None | Unset
        if isinstance(self.safety_stock_level, Unset):
            safety_stock_level = UNSET
        else:
            safety_stock_level = self.safety_stock_level

        economic_order_quantity: float | None | Unset
        if isinstance(self.economic_order_quantity, Unset):
            economic_order_quantity = UNSET
        else:
            economic_order_quantity = self.economic_order_quantity

        optimial_stock_cycle: float | None | Unset
        if isinstance(self.optimial_stock_cycle, Unset):
            optimial_stock_cycle = UNSET
        else:
            optimial_stock_cycle = self.optimial_stock_cycle

        lead_time_days: int | None | Unset
        if isinstance(self.lead_time_days, Unset):
            lead_time_days = UNSET
        else:
            lead_time_days = self.lead_time_days

        reorder_frequency_days: int | None | Unset
        if isinstance(self.reorder_frequency_days, Unset):
            reorder_frequency_days = UNSET
        else:
            reorder_frequency_days = self.reorder_frequency_days

        order_count = self.order_count

        latest_order_date: None | str | Unset
        if isinstance(self.latest_order_date, Unset):
            latest_order_date = UNSET
        elif isinstance(self.latest_order_date, datetime.datetime):
            latest_order_date = self.latest_order_date.isoformat()
        else:
            latest_order_date = self.latest_order_date

        first_purchase_date: None | str | Unset
        if isinstance(self.first_purchase_date, Unset):
            first_purchase_date = UNSET
        elif isinstance(self.first_purchase_date, datetime.datetime):
            first_purchase_date = self.first_purchase_date.isoformat()
        else:
            first_purchase_date = self.first_purchase_date

        minimum_order_quantity: int | None | Unset
        if isinstance(self.minimum_order_quantity, Unset):
            minimum_order_quantity = UNSET
        else:
            minimum_order_quantity = self.minimum_order_quantity

        batch_size: int | None | Unset
        if isinstance(self.batch_size, Unset):
            batch_size = UNSET
        else:
            batch_size = self.batch_size

        stock_on_hand: float | None | Unset
        if isinstance(self.stock_on_hand, Unset):
            stock_on_hand = UNSET
        else:
            stock_on_hand = self.stock_on_hand

        stock_on_order: float | None | Unset
        if isinstance(self.stock_on_order, Unset):
            stock_on_order = UNSET
        else:
            stock_on_order = self.stock_on_order

        finished_good_stock_on_hand: float | None | Unset
        if isinstance(self.finished_good_stock_on_hand, Unset):
            finished_good_stock_on_hand = UNSET
        else:
            finished_good_stock_on_hand = self.finished_good_stock_on_hand

        finished_good_quantity_used: float | None | Unset
        if isinstance(self.finished_good_quantity_used, Unset):
            finished_good_quantity_used = UNSET
        else:
            finished_good_quantity_used = self.finished_good_quantity_used

        component_stock_on_hand: float | None | Unset
        if isinstance(self.component_stock_on_hand, Unset):
            component_stock_on_hand = UNSET
        else:
            component_stock_on_hand = self.component_stock_on_hand

        days_until_replenishment_due: int | None | Unset
        if isinstance(self.days_until_replenishment_due, Unset):
            days_until_replenishment_due = UNSET
        else:
            days_until_replenishment_due = self.days_until_replenishment_due

        days_until_stock_out: int | None | Unset
        if isinstance(self.days_until_stock_out, Unset):
            days_until_stock_out = UNSET
        else:
            days_until_stock_out = self.days_until_stock_out

        is_uncertain = self.is_uncertain

        orders_in_previous_lead_time = self.orders_in_previous_lead_time

        avg_daily_orders_last_120_days = self.avg_daily_orders_last_120_days

        lead_demand_prediction_based_on_average: float | None | Unset
        if isinstance(self.lead_demand_prediction_based_on_average, Unset):
            lead_demand_prediction_based_on_average = UNSET
        else:
            lead_demand_prediction_based_on_average = (
                self.lead_demand_prediction_based_on_average
            )

        lead_demand_prediction_based_on_linear_regression: float | None | Unset
        if isinstance(self.lead_demand_prediction_based_on_linear_regression, Unset):
            lead_demand_prediction_based_on_linear_regression = UNSET
        else:
            lead_demand_prediction_based_on_linear_regression = (
                self.lead_demand_prediction_based_on_linear_regression
            )

        lead_demand_prediction_based_on_2_nd_order_polynomial_regression: (
            float | None | Unset
        )
        if isinstance(
            self.lead_demand_prediction_based_on_2_nd_order_polynomial_regression, Unset
        ):
            lead_demand_prediction_based_on_2_nd_order_polynomial_regression = UNSET
        else:
            lead_demand_prediction_based_on_2_nd_order_polynomial_regression = (
                self.lead_demand_prediction_based_on_2_nd_order_polynomial_regression
            )

        lead_demand_prediction_based_on_last_month: float | None | Unset
        if isinstance(self.lead_demand_prediction_based_on_last_month, Unset):
            lead_demand_prediction_based_on_last_month = UNSET
        else:
            lead_demand_prediction_based_on_last_month = (
                self.lead_demand_prediction_based_on_last_month
            )

        lead_demand_prediction_based_on_previous_leadtime_days: float | None | Unset
        if isinstance(
            self.lead_demand_prediction_based_on_previous_leadtime_days, Unset
        ):
            lead_demand_prediction_based_on_previous_leadtime_days = UNSET
        else:
            lead_demand_prediction_based_on_previous_leadtime_days = (
                self.lead_demand_prediction_based_on_previous_leadtime_days
            )

        standard_dev: float | None | Unset
        if isinstance(self.standard_dev, Unset):
            standard_dev = UNSET
        else:
            standard_dev = self.standard_dev

        service_factor: float | None | Unset
        if isinstance(self.service_factor, Unset):
            service_factor = UNSET
        else:
            service_factor = self.service_factor

        max_r_squared: float | None | Unset
        if isinstance(self.max_r_squared, Unset):
            max_r_squared = UNSET
        else:
            max_r_squared = self.max_r_squared

        min_r_squared: float | None | Unset
        if isinstance(self.min_r_squared, Unset):
            min_r_squared = UNSET
        else:
            min_r_squared = self.min_r_squared

        max_range: float | None | Unset
        if isinstance(self.max_range, Unset):
            max_range = UNSET
        else:
            max_range = self.max_range

        parent_quantity: float | None | Unset
        if isinstance(self.parent_quantity, Unset):
            parent_quantity = UNSET
        else:
            parent_quantity = self.parent_quantity

        most_accurate_algorithm_type_id: int | None | Unset
        if isinstance(self.most_accurate_algorithm_type_id, Unset):
            most_accurate_algorithm_type_id = UNSET
        else:
            most_accurate_algorithm_type_id = self.most_accurate_algorithm_type_id

        error_text: None | str | Unset
        if isinstance(self.error_text, Unset):
            error_text = UNSET
        else:
            error_text = self.error_text

        category: None | str | Unset
        if isinstance(self.category, Unset):
            category = UNSET
        else:
            category = self.category

        sub_category: None | str | Unset
        if isinstance(self.sub_category, Unset):
            sub_category = UNSET
        else:
            sub_category = self.sub_category

        brand: None | str | Unset
        if isinstance(self.brand, Unset):
            brand = UNSET
        else:
            brand = self.brand

        product_type: None | str | Unset
        if isinstance(self.product_type, Unset):
            product_type = UNSET
        else:
            product_type = self.product_type

        option1: None | str | Unset
        if isinstance(self.option1, Unset):
            option1 = UNSET
        else:
            option1 = self.option1

        option2: None | str | Unset
        if isinstance(self.option2, Unset):
            option2 = UNSET
        else:
            option2 = self.option2

        option3: None | str | Unset
        if isinstance(self.option3, Unset):
            option3 = UNSET
        else:
            option3 = self.option3

        size: None | str | Unset
        if isinstance(self.size, Unset):
            size = UNSET
        else:
            size = self.size

        product_code: None | str | Unset
        if isinstance(self.product_code, Unset):
            product_code = UNSET
        else:
            product_code = self.product_code

        name: None | str | Unset
        if isinstance(self.name, Unset):
            name = UNSET
        else:
            name = self.name

        sku_cost: float | None | Unset
        if isinstance(self.sku_cost, Unset):
            sku_cost = UNSET
        else:
            sku_cost = self.sku_cost

        sku_price: float | None | Unset
        if isinstance(self.sku_price, Unset):
            sku_price = UNSET
        else:
            sku_price = self.sku_price

        is_discontinued = self.is_discontinued

        is_unstocked = self.is_unstocked

        product_of_count = self.product_of_count

        component_of_count = self.component_of_count

        customer_count = self.customer_count

        location_count = self.location_count

        location_name: None | str | Unset
        if isinstance(self.location_name, Unset):
            location_name = UNSET
        else:
            location_name = self.location_name

        show_forecast_for_all_locations = self.show_forecast_for_all_locations

        manufacturing_time: float | None | Unset
        if isinstance(self.manufacturing_time, Unset):
            manufacturing_time = UNSET
        else:
            manufacturing_time = self.manufacturing_time

        minimum_shelf_level: float | None | Unset
        if isinstance(self.minimum_shelf_level, Unset):
            minimum_shelf_level = UNSET
        else:
            minimum_shelf_level = self.minimum_shelf_level

        maximum_shelf_level: float | None | Unset
        if isinstance(self.maximum_shelf_level, Unset):
            maximum_shelf_level = UNSET
        else:
            maximum_shelf_level = self.maximum_shelf_level

        service_level: float | None | Unset
        if isinstance(self.service_level, Unset):
            service_level = UNSET
        else:
            service_level = self.service_level

        weight: float | None | Unset
        if isinstance(self.weight, Unset):
            weight = UNSET
        else:
            weight = self.weight

        height: float | None | Unset
        if isinstance(self.height, Unset):
            height = UNSET
        else:
            height = self.height

        width: float | None | Unset
        if isinstance(self.width, Unset):
            width = UNSET
        else:
            width = self.width

        length: float | None | Unset
        if isinstance(self.length, Unset):
            length = UNSET
        else:
            length = self.length

        dimensions_cubic_meters: float | None | Unset
        if isinstance(self.dimensions_cubic_meters, Unset):
            dimensions_cubic_meters = UNSET
        else:
            dimensions_cubic_meters = self.dimensions_cubic_meters

        external_id: None | str | Unset
        if isinstance(self.external_id, Unset):
            external_id = UNSET
        else:
            external_id = self.external_id

        sku_code: None | str | Unset
        if isinstance(self.sku_code, Unset):
            sku_code = UNSET
        else:
            sku_code = self.sku_code

        external_id_parent: None | str | Unset
        if isinstance(self.external_id_parent, Unset):
            external_id_parent = UNSET
        else:
            external_id_parent = self.external_id_parent

        sku_code_parent: None | str | Unset
        if isinstance(self.sku_code_parent, Unset):
            sku_code_parent = UNSET
        else:
            sku_code_parent = self.sku_code_parent

        child_variants_count = self.child_variants_count

        field_dict: dict[str, Any] = {}

        field_dict.update({})
        if id is not UNSET:
            field_dict["id"] = id
        if sku_property_id is not UNSET:
            field_dict["skuPropertyId"] = sku_property_id
        if sku_id is not UNSET:
            field_dict["skuId"] = sku_id
        if location_id is not UNSET:
            field_dict["locationId"] = location_id
        if channel_id is not UNSET:
            field_dict["channelId"] = channel_id
        if sku_grouping_id is not UNSET:
            field_dict["skuGroupingId"] = sku_grouping_id
        if sku_optimized_results_group_id is not UNSET:
            field_dict["skuOptimizedResultsGroupId"] = sku_optimized_results_group_id
        if tenant_id is not UNSET:
            field_dict["tenantId"] = tenant_id
        if calculated_date_time is not UNSET:
            field_dict["calculatedDateTime"] = calculated_date_time
        if effective_to_date_time is not UNSET:
            field_dict["effectiveToDateTime"] = effective_to_date_time
        if reorder_point is not UNSET:
            field_dict["reorderPoint"] = reorder_point
        if order_quantity is not UNSET:
            field_dict["orderQuantity"] = order_quantity
        if lead_demand is not UNSET:
            field_dict["leadDemand"] = lead_demand
        if forecast_period_demand is not UNSET:
            field_dict["forecastPeriodDemand"] = forecast_period_demand
        if overridden_future_demand_effective_from_now is not UNSET:
            field_dict["overriddenFutureDemandEffectiveFromNow"] = (
                overridden_future_demand_effective_from_now
            )
        if safety_stock_level is not UNSET:
            field_dict["safetyStockLevel"] = safety_stock_level
        if economic_order_quantity is not UNSET:
            field_dict["economicOrderQuantity"] = economic_order_quantity
        if optimial_stock_cycle is not UNSET:
            field_dict["optimialStockCycle"] = optimial_stock_cycle
        if lead_time_days is not UNSET:
            field_dict["leadTimeDays"] = lead_time_days
        if reorder_frequency_days is not UNSET:
            field_dict["reorderFrequencyDays"] = reorder_frequency_days
        if order_count is not UNSET:
            field_dict["orderCount"] = order_count
        if latest_order_date is not UNSET:
            field_dict["latestOrderDate"] = latest_order_date
        if first_purchase_date is not UNSET:
            field_dict["firstPurchaseDate"] = first_purchase_date
        if minimum_order_quantity is not UNSET:
            field_dict["minimumOrderQuantity"] = minimum_order_quantity
        if batch_size is not UNSET:
            field_dict["batchSize"] = batch_size
        if stock_on_hand is not UNSET:
            field_dict["stockOnHand"] = stock_on_hand
        if stock_on_order is not UNSET:
            field_dict["stockOnOrder"] = stock_on_order
        if finished_good_stock_on_hand is not UNSET:
            field_dict["finishedGoodStockOnHand"] = finished_good_stock_on_hand
        if finished_good_quantity_used is not UNSET:
            field_dict["finishedGoodQuantityUsed"] = finished_good_quantity_used
        if component_stock_on_hand is not UNSET:
            field_dict["componentStockOnHand"] = component_stock_on_hand
        if days_until_replenishment_due is not UNSET:
            field_dict["daysUntilReplenishmentDue"] = days_until_replenishment_due
        if days_until_stock_out is not UNSET:
            field_dict["daysUntilStockOut"] = days_until_stock_out
        if is_uncertain is not UNSET:
            field_dict["isUncertain"] = is_uncertain
        if orders_in_previous_lead_time is not UNSET:
            field_dict["ordersInPreviousLeadTime"] = orders_in_previous_lead_time
        if avg_daily_orders_last_120_days is not UNSET:
            field_dict["avgDailyOrdersLast120Days"] = avg_daily_orders_last_120_days
        if lead_demand_prediction_based_on_average is not UNSET:
            field_dict["leadDemandPredictionBasedOnAverage"] = (
                lead_demand_prediction_based_on_average
            )
        if lead_demand_prediction_based_on_linear_regression is not UNSET:
            field_dict["leadDemandPredictionBasedOnLinearRegression"] = (
                lead_demand_prediction_based_on_linear_regression
            )
        if (
            lead_demand_prediction_based_on_2_nd_order_polynomial_regression
            is not UNSET
        ):
            field_dict["leadDemandPredictionBasedOn2ndOrderPolynomialRegression"] = (
                lead_demand_prediction_based_on_2_nd_order_polynomial_regression
            )
        if lead_demand_prediction_based_on_last_month is not UNSET:
            field_dict["leadDemandPredictionBasedOnLastMonth"] = (
                lead_demand_prediction_based_on_last_month
            )
        if lead_demand_prediction_based_on_previous_leadtime_days is not UNSET:
            field_dict["leadDemandPredictionBasedOnPreviousLeadtimeDays"] = (
                lead_demand_prediction_based_on_previous_leadtime_days
            )
        if standard_dev is not UNSET:
            field_dict["standardDev"] = standard_dev
        if service_factor is not UNSET:
            field_dict["serviceFactor"] = service_factor
        if max_r_squared is not UNSET:
            field_dict["maxRSquared"] = max_r_squared
        if min_r_squared is not UNSET:
            field_dict["minRSquared"] = min_r_squared
        if max_range is not UNSET:
            field_dict["maxRange"] = max_range
        if parent_quantity is not UNSET:
            field_dict["parentQuantity"] = parent_quantity
        if most_accurate_algorithm_type_id is not UNSET:
            field_dict["mostAccurateAlgorithmTypeId"] = most_accurate_algorithm_type_id
        if error_text is not UNSET:
            field_dict["errorText"] = error_text
        if category is not UNSET:
            field_dict["category"] = category
        if sub_category is not UNSET:
            field_dict["subCategory"] = sub_category
        if brand is not UNSET:
            field_dict["brand"] = brand
        if product_type is not UNSET:
            field_dict["productType"] = product_type
        if option1 is not UNSET:
            field_dict["option1"] = option1
        if option2 is not UNSET:
            field_dict["option2"] = option2
        if option3 is not UNSET:
            field_dict["option3"] = option3
        if size is not UNSET:
            field_dict["size"] = size
        if product_code is not UNSET:
            field_dict["productCode"] = product_code
        if name is not UNSET:
            field_dict["name"] = name
        if sku_cost is not UNSET:
            field_dict["skuCost"] = sku_cost
        if sku_price is not UNSET:
            field_dict["skuPrice"] = sku_price
        if is_discontinued is not UNSET:
            field_dict["isDiscontinued"] = is_discontinued
        if is_unstocked is not UNSET:
            field_dict["isUnstocked"] = is_unstocked
        if product_of_count is not UNSET:
            field_dict["productOfCount"] = product_of_count
        if component_of_count is not UNSET:
            field_dict["componentOfCount"] = component_of_count
        if customer_count is not UNSET:
            field_dict["customerCount"] = customer_count
        if location_count is not UNSET:
            field_dict["locationCount"] = location_count
        if location_name is not UNSET:
            field_dict["locationName"] = location_name
        if show_forecast_for_all_locations is not UNSET:
            field_dict["showForecastForAllLocations"] = show_forecast_for_all_locations
        if manufacturing_time is not UNSET:
            field_dict["manufacturingTime"] = manufacturing_time
        if minimum_shelf_level is not UNSET:
            field_dict["minimumShelfLevel"] = minimum_shelf_level
        if maximum_shelf_level is not UNSET:
            field_dict["maximumShelfLevel"] = maximum_shelf_level
        if service_level is not UNSET:
            field_dict["serviceLevel"] = service_level
        if weight is not UNSET:
            field_dict["weight"] = weight
        if height is not UNSET:
            field_dict["height"] = height
        if width is not UNSET:
            field_dict["width"] = width
        if length is not UNSET:
            field_dict["length"] = length
        if dimensions_cubic_meters is not UNSET:
            field_dict["dimensionsCubicMeters"] = dimensions_cubic_meters
        if external_id is not UNSET:
            field_dict["externalId"] = external_id
        if sku_code is not UNSET:
            field_dict["skuCode"] = sku_code
        if external_id_parent is not UNSET:
            field_dict["externalIdParent"] = external_id_parent
        if sku_code_parent is not UNSET:
            field_dict["skuCodeParent"] = sku_code_parent
        if child_variants_count is not UNSET:
            field_dict["childVariantsCount"] = child_variants_count

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        id = d.pop("id", UNSET)

        def _parse_sku_property_id(data: object) -> int | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(int | None | Unset, data)

        sku_property_id = _parse_sku_property_id(d.pop("skuPropertyId", UNSET))

        def _parse_sku_id(data: object) -> int | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(int | None | Unset, data)

        sku_id = _parse_sku_id(d.pop("skuId", UNSET))

        def _parse_location_id(data: object) -> int | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(int | None | Unset, data)

        location_id = _parse_location_id(d.pop("locationId", UNSET))

        def _parse_channel_id(data: object) -> int | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(int | None | Unset, data)

        channel_id = _parse_channel_id(d.pop("channelId", UNSET))

        def _parse_sku_grouping_id(data: object) -> int | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(int | None | Unset, data)

        sku_grouping_id = _parse_sku_grouping_id(d.pop("skuGroupingId", UNSET))

        def _parse_sku_optimized_results_group_id(data: object) -> int | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(int | None | Unset, data)

        sku_optimized_results_group_id = _parse_sku_optimized_results_group_id(
            d.pop("skuOptimizedResultsGroupId", UNSET)
        )

        def _parse_tenant_id(data: object) -> None | Unset | UUID:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                tenant_id_type_0 = UUID(data)

                return tenant_id_type_0
            except:  # noqa: E722
                pass
            return cast(None | Unset | UUID, data)

        tenant_id = _parse_tenant_id(d.pop("tenantId", UNSET))

        _calculated_date_time = d.pop("calculatedDateTime", UNSET)
        calculated_date_time: datetime.datetime | Unset
        if isinstance(_calculated_date_time, Unset):
            calculated_date_time = UNSET
        else:
            calculated_date_time = isoparse(_calculated_date_time)

        def _parse_effective_to_date_time(
            data: object,
        ) -> datetime.datetime | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                effective_to_date_time_type_0 = isoparse(data)

                return effective_to_date_time_type_0
            except:  # noqa: E722
                pass
            return cast(datetime.datetime | None | Unset, data)

        effective_to_date_time = _parse_effective_to_date_time(
            d.pop("effectiveToDateTime", UNSET)
        )

        def _parse_reorder_point(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        reorder_point = _parse_reorder_point(d.pop("reorderPoint", UNSET))

        def _parse_order_quantity(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        order_quantity = _parse_order_quantity(d.pop("orderQuantity", UNSET))

        def _parse_lead_demand(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        lead_demand = _parse_lead_demand(d.pop("leadDemand", UNSET))

        def _parse_forecast_period_demand(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        forecast_period_demand = _parse_forecast_period_demand(
            d.pop("forecastPeriodDemand", UNSET)
        )

        def _parse_overridden_future_demand_effective_from_now(
            data: object,
        ) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        overridden_future_demand_effective_from_now = (
            _parse_overridden_future_demand_effective_from_now(
                d.pop("overriddenFutureDemandEffectiveFromNow", UNSET)
            )
        )

        def _parse_safety_stock_level(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        safety_stock_level = _parse_safety_stock_level(d.pop("safetyStockLevel", UNSET))

        def _parse_economic_order_quantity(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        economic_order_quantity = _parse_economic_order_quantity(
            d.pop("economicOrderQuantity", UNSET)
        )

        def _parse_optimial_stock_cycle(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        optimial_stock_cycle = _parse_optimial_stock_cycle(
            d.pop("optimialStockCycle", UNSET)
        )

        def _parse_lead_time_days(data: object) -> int | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(int | None | Unset, data)

        lead_time_days = _parse_lead_time_days(d.pop("leadTimeDays", UNSET))

        def _parse_reorder_frequency_days(data: object) -> int | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(int | None | Unset, data)

        reorder_frequency_days = _parse_reorder_frequency_days(
            d.pop("reorderFrequencyDays", UNSET)
        )

        order_count = d.pop("orderCount", UNSET)

        def _parse_latest_order_date(data: object) -> datetime.datetime | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                latest_order_date_type_0 = isoparse(data)

                return latest_order_date_type_0
            except:  # noqa: E722
                pass
            return cast(datetime.datetime | None | Unset, data)

        latest_order_date = _parse_latest_order_date(d.pop("latestOrderDate", UNSET))

        def _parse_first_purchase_date(
            data: object,
        ) -> datetime.datetime | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                first_purchase_date_type_0 = isoparse(data)

                return first_purchase_date_type_0
            except:  # noqa: E722
                pass
            return cast(datetime.datetime | None | Unset, data)

        first_purchase_date = _parse_first_purchase_date(
            d.pop("firstPurchaseDate", UNSET)
        )

        def _parse_minimum_order_quantity(data: object) -> int | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(int | None | Unset, data)

        minimum_order_quantity = _parse_minimum_order_quantity(
            d.pop("minimumOrderQuantity", UNSET)
        )

        def _parse_batch_size(data: object) -> int | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(int | None | Unset, data)

        batch_size = _parse_batch_size(d.pop("batchSize", UNSET))

        def _parse_stock_on_hand(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        stock_on_hand = _parse_stock_on_hand(d.pop("stockOnHand", UNSET))

        def _parse_stock_on_order(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        stock_on_order = _parse_stock_on_order(d.pop("stockOnOrder", UNSET))

        def _parse_finished_good_stock_on_hand(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        finished_good_stock_on_hand = _parse_finished_good_stock_on_hand(
            d.pop("finishedGoodStockOnHand", UNSET)
        )

        def _parse_finished_good_quantity_used(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        finished_good_quantity_used = _parse_finished_good_quantity_used(
            d.pop("finishedGoodQuantityUsed", UNSET)
        )

        def _parse_component_stock_on_hand(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        component_stock_on_hand = _parse_component_stock_on_hand(
            d.pop("componentStockOnHand", UNSET)
        )

        def _parse_days_until_replenishment_due(data: object) -> int | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(int | None | Unset, data)

        days_until_replenishment_due = _parse_days_until_replenishment_due(
            d.pop("daysUntilReplenishmentDue", UNSET)
        )

        def _parse_days_until_stock_out(data: object) -> int | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(int | None | Unset, data)

        days_until_stock_out = _parse_days_until_stock_out(
            d.pop("daysUntilStockOut", UNSET)
        )

        is_uncertain = d.pop("isUncertain", UNSET)

        orders_in_previous_lead_time = d.pop("ordersInPreviousLeadTime", UNSET)

        avg_daily_orders_last_120_days = d.pop("avgDailyOrdersLast120Days", UNSET)

        def _parse_lead_demand_prediction_based_on_average(
            data: object,
        ) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        lead_demand_prediction_based_on_average = (
            _parse_lead_demand_prediction_based_on_average(
                d.pop("leadDemandPredictionBasedOnAverage", UNSET)
            )
        )

        def _parse_lead_demand_prediction_based_on_linear_regression(
            data: object,
        ) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        lead_demand_prediction_based_on_linear_regression = (
            _parse_lead_demand_prediction_based_on_linear_regression(
                d.pop("leadDemandPredictionBasedOnLinearRegression", UNSET)
            )
        )

        def _parse_lead_demand_prediction_based_on_2_nd_order_polynomial_regression(
            data: object,
        ) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        lead_demand_prediction_based_on_2_nd_order_polynomial_regression = (
            _parse_lead_demand_prediction_based_on_2_nd_order_polynomial_regression(
                d.pop("leadDemandPredictionBasedOn2ndOrderPolynomialRegression", UNSET)
            )
        )

        def _parse_lead_demand_prediction_based_on_last_month(
            data: object,
        ) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        lead_demand_prediction_based_on_last_month = (
            _parse_lead_demand_prediction_based_on_last_month(
                d.pop("leadDemandPredictionBasedOnLastMonth", UNSET)
            )
        )

        def _parse_lead_demand_prediction_based_on_previous_leadtime_days(
            data: object,
        ) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        lead_demand_prediction_based_on_previous_leadtime_days = (
            _parse_lead_demand_prediction_based_on_previous_leadtime_days(
                d.pop("leadDemandPredictionBasedOnPreviousLeadtimeDays", UNSET)
            )
        )

        def _parse_standard_dev(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        standard_dev = _parse_standard_dev(d.pop("standardDev", UNSET))

        def _parse_service_factor(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        service_factor = _parse_service_factor(d.pop("serviceFactor", UNSET))

        def _parse_max_r_squared(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        max_r_squared = _parse_max_r_squared(d.pop("maxRSquared", UNSET))

        def _parse_min_r_squared(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        min_r_squared = _parse_min_r_squared(d.pop("minRSquared", UNSET))

        def _parse_max_range(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        max_range = _parse_max_range(d.pop("maxRange", UNSET))

        def _parse_parent_quantity(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        parent_quantity = _parse_parent_quantity(d.pop("parentQuantity", UNSET))

        def _parse_most_accurate_algorithm_type_id(data: object) -> int | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(int | None | Unset, data)

        most_accurate_algorithm_type_id = _parse_most_accurate_algorithm_type_id(
            d.pop("mostAccurateAlgorithmTypeId", UNSET)
        )

        def _parse_error_text(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        error_text = _parse_error_text(d.pop("errorText", UNSET))

        def _parse_category(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        category = _parse_category(d.pop("category", UNSET))

        def _parse_sub_category(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        sub_category = _parse_sub_category(d.pop("subCategory", UNSET))

        def _parse_brand(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        brand = _parse_brand(d.pop("brand", UNSET))

        def _parse_product_type(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        product_type = _parse_product_type(d.pop("productType", UNSET))

        def _parse_option1(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        option1 = _parse_option1(d.pop("option1", UNSET))

        def _parse_option2(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        option2 = _parse_option2(d.pop("option2", UNSET))

        def _parse_option3(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        option3 = _parse_option3(d.pop("option3", UNSET))

        def _parse_size(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        size = _parse_size(d.pop("size", UNSET))

        def _parse_product_code(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        product_code = _parse_product_code(d.pop("productCode", UNSET))

        def _parse_name(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        name = _parse_name(d.pop("name", UNSET))

        def _parse_sku_cost(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        sku_cost = _parse_sku_cost(d.pop("skuCost", UNSET))

        def _parse_sku_price(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        sku_price = _parse_sku_price(d.pop("skuPrice", UNSET))

        is_discontinued = d.pop("isDiscontinued", UNSET)

        is_unstocked = d.pop("isUnstocked", UNSET)

        product_of_count = d.pop("productOfCount", UNSET)

        component_of_count = d.pop("componentOfCount", UNSET)

        customer_count = d.pop("customerCount", UNSET)

        location_count = d.pop("locationCount", UNSET)

        def _parse_location_name(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        location_name = _parse_location_name(d.pop("locationName", UNSET))

        show_forecast_for_all_locations = d.pop("showForecastForAllLocations", UNSET)

        def _parse_manufacturing_time(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        manufacturing_time = _parse_manufacturing_time(
            d.pop("manufacturingTime", UNSET)
        )

        def _parse_minimum_shelf_level(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        minimum_shelf_level = _parse_minimum_shelf_level(
            d.pop("minimumShelfLevel", UNSET)
        )

        def _parse_maximum_shelf_level(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        maximum_shelf_level = _parse_maximum_shelf_level(
            d.pop("maximumShelfLevel", UNSET)
        )

        def _parse_service_level(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        service_level = _parse_service_level(d.pop("serviceLevel", UNSET))

        def _parse_weight(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        weight = _parse_weight(d.pop("weight", UNSET))

        def _parse_height(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        height = _parse_height(d.pop("height", UNSET))

        def _parse_width(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        width = _parse_width(d.pop("width", UNSET))

        def _parse_length(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        length = _parse_length(d.pop("length", UNSET))

        def _parse_dimensions_cubic_meters(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        dimensions_cubic_meters = _parse_dimensions_cubic_meters(
            d.pop("dimensionsCubicMeters", UNSET)
        )

        def _parse_external_id(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        external_id = _parse_external_id(d.pop("externalId", UNSET))

        def _parse_sku_code(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        sku_code = _parse_sku_code(d.pop("skuCode", UNSET))

        def _parse_external_id_parent(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        external_id_parent = _parse_external_id_parent(d.pop("externalIdParent", UNSET))

        def _parse_sku_code_parent(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        sku_code_parent = _parse_sku_code_parent(d.pop("skuCodeParent", UNSET))

        child_variants_count = d.pop("childVariantsCount", UNSET)

        sku_optimized_results_dto = cls(
            id=id,
            sku_property_id=sku_property_id,
            sku_id=sku_id,
            location_id=location_id,
            channel_id=channel_id,
            sku_grouping_id=sku_grouping_id,
            sku_optimized_results_group_id=sku_optimized_results_group_id,
            tenant_id=tenant_id,
            calculated_date_time=calculated_date_time,
            effective_to_date_time=effective_to_date_time,
            reorder_point=reorder_point,
            order_quantity=order_quantity,
            lead_demand=lead_demand,
            forecast_period_demand=forecast_period_demand,
            overridden_future_demand_effective_from_now=overridden_future_demand_effective_from_now,
            safety_stock_level=safety_stock_level,
            economic_order_quantity=economic_order_quantity,
            optimial_stock_cycle=optimial_stock_cycle,
            lead_time_days=lead_time_days,
            reorder_frequency_days=reorder_frequency_days,
            order_count=order_count,
            latest_order_date=latest_order_date,
            first_purchase_date=first_purchase_date,
            minimum_order_quantity=minimum_order_quantity,
            batch_size=batch_size,
            stock_on_hand=stock_on_hand,
            stock_on_order=stock_on_order,
            finished_good_stock_on_hand=finished_good_stock_on_hand,
            finished_good_quantity_used=finished_good_quantity_used,
            component_stock_on_hand=component_stock_on_hand,
            days_until_replenishment_due=days_until_replenishment_due,
            days_until_stock_out=days_until_stock_out,
            is_uncertain=is_uncertain,
            orders_in_previous_lead_time=orders_in_previous_lead_time,
            avg_daily_orders_last_120_days=avg_daily_orders_last_120_days,
            lead_demand_prediction_based_on_average=lead_demand_prediction_based_on_average,
            lead_demand_prediction_based_on_linear_regression=lead_demand_prediction_based_on_linear_regression,
            lead_demand_prediction_based_on_2_nd_order_polynomial_regression=lead_demand_prediction_based_on_2_nd_order_polynomial_regression,
            lead_demand_prediction_based_on_last_month=lead_demand_prediction_based_on_last_month,
            lead_demand_prediction_based_on_previous_leadtime_days=lead_demand_prediction_based_on_previous_leadtime_days,
            standard_dev=standard_dev,
            service_factor=service_factor,
            max_r_squared=max_r_squared,
            min_r_squared=min_r_squared,
            max_range=max_range,
            parent_quantity=parent_quantity,
            most_accurate_algorithm_type_id=most_accurate_algorithm_type_id,
            error_text=error_text,
            category=category,
            sub_category=sub_category,
            brand=brand,
            product_type=product_type,
            option1=option1,
            option2=option2,
            option3=option3,
            size=size,
            product_code=product_code,
            name=name,
            sku_cost=sku_cost,
            sku_price=sku_price,
            is_discontinued=is_discontinued,
            is_unstocked=is_unstocked,
            product_of_count=product_of_count,
            component_of_count=component_of_count,
            customer_count=customer_count,
            location_count=location_count,
            location_name=location_name,
            show_forecast_for_all_locations=show_forecast_for_all_locations,
            manufacturing_time=manufacturing_time,
            minimum_shelf_level=minimum_shelf_level,
            maximum_shelf_level=maximum_shelf_level,
            service_level=service_level,
            weight=weight,
            height=height,
            width=width,
            length=length,
            dimensions_cubic_meters=dimensions_cubic_meters,
            external_id=external_id,
            sku_code=sku_code,
            external_id_parent=external_id_parent,
            sku_code_parent=sku_code_parent,
            child_variants_count=child_variants_count,
        )

        return sku_optimized_results_dto



"""Sales Order management tools for StockTrim MCP Server."""

from __future__ import annotations

import logging
from datetime import datetime

from fastmcp import Context, FastMCP
from pydantic import BaseModel, Field

from stocktrim_mcp_server.dependencies import get_services
from stocktrim_mcp_server.utils import unset_to_none

logger = logging.getLogger(__name__)

# ============================================================================
# Tool 1: create_sales_order
# ============================================================================


class CreateSalesOrderRequest(BaseModel):
    """Request model for creating a sales order."""

    product_id: str = Field(..., description="Product ID for the order")
    order_date: datetime = Field(..., description="Order date (ISO format)")
    quantity: float = Field(..., gt=0, description="Quantity ordered (must be > 0)")
    external_reference_id: str | None = Field(
        None, description="External reference ID (optional)"
    )
    unit_price: float | None = Field(None, ge=0, description="Unit price (optional)")
    location_code: str | None = Field(None, description="Location code (optional)")
    location_name: str | None = Field(None, description="Location name (optional)")
    customer_code: str | None = Field(None, description="Customer code (optional)")
    customer_name: str | None = Field(None, description="Customer name (optional)")


class SalesOrderInfo(BaseModel):
    """Sales order information."""

    id: int | None
    product_id: str
    order_date: datetime
    quantity: float
    external_reference_id: str | None
    unit_price: float | None
    location_code: str | None
    location_name: str | None
    customer_code: str | None
    customer_name: str | None
    location_id: int | None


async def _create_sales_order_impl(
    request: CreateSalesOrderRequest, context: Context
) -> SalesOrderInfo:
    """Implementation of create_sales_order tool.

    Args:
        request: Request containing sales order details
        context: Server context with StockTrimClient

    Returns:
        SalesOrderInfo with created order details

    Raises:
        Exception: If API call fails
    """
    services = get_services(context)

    # Create the sales order via service
    result = await services.sales_orders.create(
        product_id=request.product_id,
        order_date=request.order_date,
        quantity=request.quantity,
        external_reference_id=request.external_reference_id,
        unit_price=request.unit_price,
        location_code=request.location_code,
        location_name=request.location_name,
        customer_code=request.customer_code,
        customer_name=request.customer_name,
    )

    # Build response model (convert UNSET to None for Pydantic)
    return SalesOrderInfo(
        id=unset_to_none(result.id),
        product_id=result.product_id,
        order_date=result.order_date,
        quantity=result.quantity,
        external_reference_id=unset_to_none(result.external_reference_id),
        unit_price=unset_to_none(result.unit_price),
        location_code=unset_to_none(result.location_code),
        location_name=unset_to_none(result.location_name),
        customer_code=unset_to_none(result.customer_code),
        customer_name=unset_to_none(result.customer_name),
        location_id=unset_to_none(result.location_id),
    )


async def create_sales_order(
    request: CreateSalesOrderRequest, context: Context
) -> SalesOrderInfo:
    """Create a new sales order.

    This tool creates a sales order in StockTrim for a specific product.
    Note: StockTrim sales orders are product-based (one product per order).

    Args:
        request: Request containing sales order details
        context: Server context with StockTrimClient

    Returns:
        SalesOrderInfo with created order details

    Example:
        Request: {
            "product_id": "WIDGET-001",
            "order_date": "2024-01-15T10:00:00Z",
            "quantity": 10.0,
            "customer_code": "CUST-001",
            "unit_price": 29.99
        }
        Returns: {
            "id": 123,
            "product_id": "WIDGET-001",
            "quantity": 10.0,
            ...
        }
    """
    return await _create_sales_order_impl(request, context)


# ============================================================================
# Tool 2: get_sales_orders
# ============================================================================


class GetSalesOrdersRequest(BaseModel):
    """Request model for getting sales orders."""

    product_id: str | None = Field(None, description="Filter by product ID (optional)")


class GetSalesOrdersResponse(BaseModel):
    """Response containing sales orders."""

    sales_orders: list[SalesOrderInfo]
    total_count: int


async def _get_sales_orders_impl(
    request: GetSalesOrdersRequest, context: Context
) -> GetSalesOrdersResponse:
    """Implementation of get_sales_orders tool.

    Args:
        request: Request with optional product_id filter
        context: Server context with StockTrimClient

    Returns:
        GetSalesOrdersResponse with sales orders

    Raises:
        Exception: If API call fails
    """
    services = get_services(context)

    # Get sales orders via service
    orders = await services.sales_orders.get_all(product_id=request.product_id)

    # Build response (convert UNSET to None for Pydantic)
    order_infos = [
        SalesOrderInfo(
            id=unset_to_none(order.id),
            product_id=order.product_id,
            order_date=order.order_date,
            quantity=order.quantity,
            external_reference_id=unset_to_none(order.external_reference_id),
            unit_price=unset_to_none(order.unit_price),
            location_code=unset_to_none(order.location_code),
            location_name=unset_to_none(order.location_name),
            customer_code=unset_to_none(order.customer_code),
            customer_name=unset_to_none(order.customer_name),
            location_id=unset_to_none(order.location_id),
        )
        for order in orders
    ]

    return GetSalesOrdersResponse(
        sales_orders=order_infos,
        total_count=len(order_infos),
    )


async def get_sales_orders(
    request: GetSalesOrdersRequest, context: Context
) -> GetSalesOrdersResponse:
    """Get sales orders, optionally filtered by product.

    This tool retrieves sales orders from StockTrim. You can optionally
    filter by product ID to see orders for a specific product.

    Args:
        request: Request with optional product_id filter
        context: Server context with StockTrimClient

    Returns:
        GetSalesOrdersResponse with sales orders

    Example:
        Request: {"product_id": "WIDGET-001"}
        Returns: {"sales_orders": [...], "total_count": 5}

        Request: {}
        Returns: {"sales_orders": [...], "total_count": 50}
    """
    return await _get_sales_orders_impl(request, context)


# ============================================================================
# Tool 3: list_sales_orders (alias for backward compatibility)
# ============================================================================


class ListSalesOrdersRequest(BaseModel):
    """Request model for listing sales orders (alias for get_sales_orders)."""

    product_id: str | None = Field(None, description="Filter by product ID (optional)")


class ListSalesOrdersResponse(BaseModel):
    """Response containing sales orders (alias for get_sales_orders)."""

    sales_orders: list[SalesOrderInfo]
    total_count: int


async def list_sales_orders(
    request: ListSalesOrdersRequest, context: Context
) -> ListSalesOrdersResponse:
    """List all sales orders with optional product filter.

    This is an alias for get_sales_orders for backward compatibility.

    Args:
        request: Request with optional product_id filter
        context: Server context with StockTrimClient

    Returns:
        ListSalesOrdersResponse with sales orders

    Example:
        Request: {}
        Returns: {"sales_orders": [...], "total_count": 50}
    """
    get_request = GetSalesOrdersRequest(product_id=request.product_id)
    get_response = await _get_sales_orders_impl(get_request, context)

    return ListSalesOrdersResponse(
        sales_orders=get_response.sales_orders,
        total_count=get_response.total_count,
    )


# ============================================================================
# Tool 4: delete_sales_orders
# ============================================================================


class DeleteSalesOrdersRequest(BaseModel):
    """Request model for deleting sales orders."""

    product_id: str | None = Field(
        None,
        description="Product ID to filter deletions (deletes all orders for this product)",
    )


class DeleteSalesOrdersResponse(BaseModel):
    """Response for sales order deletion."""

    success: bool
    message: str


async def _delete_sales_orders_impl(
    request: DeleteSalesOrdersRequest, context: Context
) -> DeleteSalesOrdersResponse:
    """Implementation of delete_sales_orders tool.

    Args:
        request: Request with optional product_id filter
        context: Server context with StockTrimClient

    Returns:
        DeleteSalesOrdersResponse indicating success

    Raises:
        ValueError: If no filter provided (safety measure)
        Exception: If API call fails
    """
    if not request.product_id:
        # Safety measure: require a filter to avoid deleting all orders
        raise ValueError(
            "product_id is required for deletion to prevent accidental bulk deletion."
        )

    services = get_services(context)

    # Delete sales orders via service
    success, message = await services.sales_orders.delete_for_product(
        request.product_id
    )

    return DeleteSalesOrdersResponse(
        success=success,
        message=message,
    )


async def delete_sales_orders(
    request: DeleteSalesOrdersRequest, context: Context
) -> DeleteSalesOrdersResponse:
    """Delete sales orders for a specific product.

    This tool deletes all sales orders associated with a product.
    For safety, product_id is required (cannot delete all orders without filter).

    Args:
        request: Request with product_id filter
        context: Server context with StockTrimClient

    Returns:
        DeleteSalesOrdersResponse indicating success

    Example:
        Request: {"product_id": "WIDGET-001"}
        Returns: {
            "success": true,
            "message": "Sales orders for product WIDGET-001 deleted successfully",
            "deleted_count": 5
        }
    """
    return await _delete_sales_orders_impl(request, context)


# ============================================================================
# Tool Registration
# ============================================================================


def register_tools(mcp: FastMCP) -> None:
    """Register sales order tools with FastMCP server.

    Args:
        mcp: FastMCP server instance
    """
    mcp.tool()(create_sales_order)
    mcp.tool()(get_sales_orders)
    mcp.tool()(list_sales_orders)
    mcp.tool()(delete_sales_orders)



"""Supplier management tools for StockTrim MCP Server."""

from __future__ import annotations

import logging

from fastmcp import Context, FastMCP
from pydantic import BaseModel, Field

from stocktrim_mcp_server.dependencies import get_services

logger = logging.getLogger(__name__)

# ============================================================================
# Tool 1: get_supplier
# ============================================================================


class GetSupplierRequest(BaseModel):
    """Request model for getting a supplier."""

    code: str = Field(..., description="Supplier code to retrieve")


class SupplierInfo(BaseModel):
    """Supplier information."""

    code: str
    name: str | None
    email: str | None
    primary_contact: str | None


async def get_supplier(
    request: GetSupplierRequest, context: Context
) -> SupplierInfo | None:
    """Get a supplier by code.

    This tool retrieves detailed information about a specific supplier
    from StockTrim.

    Args:
        request: Request containing supplier code
        context: Server context with StockTrimClient

    Returns:
        SupplierInfo if found, None if not found

    Example:
        Request: {"code": "SUP-001"}
        Returns: {"code": "SUP-001", "name": "Acme Supplies", ...}
    """
    services = get_services(context)
    supplier = await services.suppliers.get_by_code(request.code)

    if not supplier:
        return None

    # Build SupplierInfo from response
    return SupplierInfo(
        code=supplier.supplier_code,
        name=supplier.supplier_name,
        email=supplier.email_address,
        primary_contact=supplier.primary_contact_name,
    )


# ============================================================================
# Tool 2: list_suppliers
# ============================================================================


class ListSuppliersRequest(BaseModel):
    """Request model for listing suppliers."""

    active_only: bool = Field(
        default=False, description="Only return active suppliers (default: false)"
    )


class ListSuppliersResponse(BaseModel):
    """Response containing suppliers."""

    suppliers: list[SupplierInfo]
    total_count: int


async def list_suppliers(
    request: ListSuppliersRequest, context: Context
) -> ListSuppliersResponse:
    """List all suppliers.

    This tool retrieves all suppliers from StockTrim,
    optionally filtered by active status.

    Args:
        request: Request with filter options
        context: Server context with StockTrimClient

    Returns:
        ListSuppliersResponse with suppliers

    Example:
        Request: {"active_only": true}
        Returns: {"suppliers": [...], "total_count": 10}
    """
    services = get_services(context)
    suppliers = await services.suppliers.list_all(request.active_only)

    # Build response
    supplier_infos = [
        SupplierInfo(
            code=s.supplier_code,
            name=s.supplier_name,
            email=s.email_address,
            primary_contact=s.primary_contact_name,
        )
        for s in suppliers
    ]

    return ListSuppliersResponse(
        suppliers=supplier_infos,
        total_count=len(supplier_infos),
    )


# ============================================================================
# Tool 3: create_supplier
# ============================================================================


class CreateSupplierRequest(BaseModel):
    """Request model for creating a supplier."""

    code: str = Field(..., description="Unique supplier code")
    name: str = Field(..., description="Supplier name")
    email: str | None = Field(default=None, description="Supplier email")
    primary_contact: str | None = Field(
        default=None, description="Primary contact name"
    )


async def create_supplier(
    request: CreateSupplierRequest, context: Context
) -> SupplierInfo:
    """Create a new supplier.

    This tool creates a new supplier in StockTrim.

    Args:
        request: Request containing supplier details
        context: Server context with StockTrimClient

    Returns:
        SupplierInfo for the created supplier

    Example:
        Request: {"code": "SUP-001", "name": "Acme Supplies", "email": "contact@acme.com"}
        Returns: {"code": "SUP-001", "name": "Acme Supplies", ...}
    """
    services = get_services(context)
    created_supplier = await services.suppliers.create(
        code=request.code,
        name=request.name,
        email=request.email,
        primary_contact=request.primary_contact,
    )

    # Build SupplierInfo from response
    return SupplierInfo(
        code=created_supplier.supplier_code,
        name=created_supplier.supplier_name,
        email=created_supplier.email_address,
        primary_contact=created_supplier.primary_contact_name,
    )


# ============================================================================
# Tool 4: delete_supplier
# ============================================================================


class DeleteSupplierRequest(BaseModel):
    """Request model for deleting a supplier."""

    code: str = Field(..., description="Supplier code to delete")


class DeleteSupplierResponse(BaseModel):
    """Response for supplier deletion."""

    success: bool
    message: str


async def delete_supplier(
    request: DeleteSupplierRequest, context: Context
) -> DeleteSupplierResponse:
    """Delete a supplier by code.

    This tool deletes a supplier from StockTrim.

    Args:
        request: Request containing supplier code
        context: Server context with StockTrimClient

    Returns:
        DeleteSupplierResponse indicating success

    Example:
        Request: {"code": "SUP-001"}
        Returns: {"success": true, "message": "Supplier SUP-001 deleted successfully"}
    """
    services = get_services(context)
    success, message = await services.suppliers.delete(request.code)

    return DeleteSupplierResponse(
        success=success,
        message=message,
    )


# ============================================================================
# Tool Registration
# ============================================================================


def register_tools(mcp: FastMCP) -> None:
    """Register supplier tools with FastMCP server.

    Args:
        mcp: FastMCP server instance
    """
    mcp.tool()(get_supplier)
    mcp.tool()(list_suppliers)
    mcp.tool()(create_supplier)
    mcp.tool()(delete_supplier)



"""Tests for sales order management foundation tools."""

from datetime import datetime
from unittest.mock import AsyncMock

import pytest
from pydantic import ValidationError

from stocktrim_mcp_server.tools.foundation.sales_orders import (
    CreateSalesOrderRequest,
    DeleteSalesOrdersRequest,
    GetSalesOrdersRequest,
    ListSalesOrdersRequest,
    create_sales_order,
    delete_sales_orders,
    get_sales_orders,
    list_sales_orders,
)
from stocktrim_public_api_client.generated.models.sales_order_response_dto import (
    SalesOrderResponseDto,
)


@pytest.fixture
def sample_sales_order():
    """Create a sample sales order for testing."""
    return SalesOrderResponseDto(
        id=789,
        product_id="prod-123",
        order_date=datetime(2024, 1, 15, 10, 0, 0),
        quantity=10.0,
        external_reference_id="SO-2024-001",
        unit_price=29.99,
        location_code="WAREHOUSE-A",
        location_name="Main Warehouse",
        customer_code="CUST-001",
        customer_name="Test Customer",
        location_id=1,
    )


@pytest.fixture
def extended_mock_context(mock_context):
    """Extend mock context with sales_orders service."""
    from stocktrim_mcp_server.services.sales_orders import SalesOrderService

    mock_service = AsyncMock(spec=SalesOrderService)
    mock_context.request_context.lifespan_context.sales_orders = mock_service
    return mock_context


# ============================================================================
# Test create_sales_order
# ============================================================================


@pytest.mark.asyncio
async def test_create_sales_order_success(extended_mock_context, sample_sales_order):
    """Test successfully creating a sales order."""
    # Setup
    mock_service = extended_mock_context.request_context.lifespan_context.sales_orders
    mock_service.create.return_value = sample_sales_order

    # Execute
    request = CreateSalesOrderRequest(
        product_id="prod-123",
        order_date=datetime(2024, 1, 15, 10, 0, 0),
        quantity=10.0,
        customer_code="CUST-001",
        unit_price=29.99,
    )
    response = await create_sales_order(request, extended_mock_context)

    # Verify
    assert response.id == 789
    assert response.product_id == "prod-123"
    assert response.quantity == 10.0
    assert response.customer_code == "CUST-001"
    assert response.unit_price == 29.99
    mock_service.create.assert_called_once()


@pytest.mark.asyncio
async def test_create_sales_order_minimal(extended_mock_context, sample_sales_order):
    """Test creating a sales order with minimal required fields."""
    # Setup
    mock_service = extended_mock_context.request_context.lifespan_context.sales_orders
    minimal_order = SalesOrderResponseDto(
        id=790,
        product_id="prod-456",
        order_date=datetime(2024, 1, 15, 10, 0, 0),
        quantity=5.0,
    )
    mock_service.create.return_value = minimal_order

    # Execute
    request = CreateSalesOrderRequest(
        product_id="prod-456",
        order_date=datetime(2024, 1, 15, 10, 0, 0),
        quantity=5.0,
    )
    response = await create_sales_order(request, extended_mock_context)

    # Verify
    assert response.id == 790
    assert response.product_id == "prod-456"
    assert response.quantity == 5.0
    assert response.customer_code is None
    assert response.unit_price is None


@pytest.mark.asyncio
async def test_create_sales_order_empty_product_id(extended_mock_context):
    """Test creating a sales order with empty product_id raises error."""
    # Setup - service layer validates and raises ValueError
    mock_service = extended_mock_context.request_context.lifespan_context.sales_orders
    mock_service.create.side_effect = ValueError("Product ID cannot be empty")

    # Execute and verify
    request = CreateSalesOrderRequest(
        product_id="",
        order_date=datetime(2024, 1, 15, 10, 0, 0),
        quantity=10.0,
    )
    with pytest.raises(ValueError, match="Product ID cannot be empty"):
        await create_sales_order(request, extended_mock_context)


@pytest.mark.asyncio
async def test_create_sales_order_zero_quantity(extended_mock_context):
    """Test creating a sales order with zero quantity raises error."""
    # Pydantic validation prevents zero quantity at the request level
    with pytest.raises(ValidationError):
        CreateSalesOrderRequest(
            product_id="prod-123",
            order_date=datetime(2024, 1, 15, 10, 0, 0),
            quantity=0.0,
        )


# ============================================================================
# Test get_sales_orders
# ============================================================================


@pytest.mark.asyncio
async def test_get_sales_orders_all(extended_mock_context, sample_sales_order):
    """Test getting all sales orders."""
    # Setup
    mock_service = extended_mock_context.request_context.lifespan_context.sales_orders
    mock_service.get_all.return_value = [
        sample_sales_order,
        SalesOrderResponseDto(
            id=790,
            product_id="prod-456",
            order_date=datetime(2024, 1, 16, 10, 0, 0),
            quantity=5.0,
        ),
    ]

    # Execute
    request = GetSalesOrdersRequest()
    response = await get_sales_orders(request, extended_mock_context)

    # Verify
    assert response.total_count == 2
    assert len(response.sales_orders) == 2
    assert response.sales_orders[0].id == 789
    assert response.sales_orders[1].id == 790
    mock_service.get_all.assert_called_once_with(product_id=None)


@pytest.mark.asyncio
async def test_get_sales_orders_by_product(extended_mock_context, sample_sales_order):
    """Test getting sales orders filtered by product ID."""
    # Setup
    mock_service = extended_mock_context.request_context.lifespan_context.sales_orders
    mock_service.get_all.return_value = [sample_sales_order]

    # Execute
    request = GetSalesOrdersRequest(product_id="prod-123")
    response = await get_sales_orders(request, extended_mock_context)

    # Verify
    assert response.total_count == 1
    assert len(response.sales_orders) == 1
    assert response.sales_orders[0].product_id == "prod-123"
    mock_service.get_all.assert_called_once_with(product_id="prod-123")


@pytest.mark.asyncio
async def test_get_sales_orders_empty_list(extended_mock_context):
    """Test getting sales orders when none exist."""
    # Setup
    mock_service = extended_mock_context.request_context.lifespan_context.sales_orders
    mock_service.get_all.return_value = []

    # Execute
    request = GetSalesOrdersRequest()
    response = await get_sales_orders(request, extended_mock_context)

    # Verify
    assert response.total_count == 0
    assert len(response.sales_orders) == 0


@pytest.mark.asyncio
async def test_get_sales_orders_single_object(
    extended_mock_context, sample_sales_order
):
    """Test getting sales orders when API returns single object instead of list."""
    # Setup - service layer already handles this, so it returns a list
    mock_service = extended_mock_context.request_context.lifespan_context.sales_orders
    mock_service.get_all.return_value = [sample_sales_order]

    # Execute
    request = GetSalesOrdersRequest()
    response = await get_sales_orders(request, extended_mock_context)

    # Verify
    assert response.total_count == 1
    assert len(response.sales_orders) == 1
    assert response.sales_orders[0].id == 789


# ============================================================================
# Test list_sales_orders (alias)
# ============================================================================


@pytest.mark.asyncio
async def test_list_sales_orders_all(extended_mock_context, sample_sales_order):
    """Test listing all sales orders using alias."""
    # Setup
    mock_service = extended_mock_context.request_context.lifespan_context.sales_orders
    mock_service.get_all.return_value = [sample_sales_order]

    # Execute
    request = ListSalesOrdersRequest()
    response = await list_sales_orders(request, extended_mock_context)

    # Verify
    assert response.total_count == 1
    assert len(response.sales_orders) == 1
    mock_service.get_all.assert_called_once()


@pytest.mark.asyncio
async def test_list_sales_orders_by_product(extended_mock_context, sample_sales_order):
    """Test listing sales orders filtered by product using alias."""
    # Setup
    mock_service = extended_mock_context.request_context.lifespan_context.sales_orders
    mock_service.get_all.return_value = [sample_sales_order]

    # Execute
    request = ListSalesOrdersRequest(product_id="prod-123")
    response = await list_sales_orders(request, extended_mock_context)

    # Verify
    assert response.total_count == 1
    assert response.sales_orders[0].product_id == "prod-123"


# ============================================================================
# Test delete_sales_orders
# ============================================================================


@pytest.mark.asyncio
async def test_delete_sales_orders_by_product(
    extended_mock_context, sample_sales_order
):
    """Test deleting sales orders for a specific product."""
    # Setup
    mock_service = extended_mock_context.request_context.lifespan_context.sales_orders
    mock_service.delete_for_product.return_value = (
        True,
        "Sales orders for product prod-123 deleted successfully",
    )

    # Execute
    request = DeleteSalesOrdersRequest(product_id="prod-123")
    response = await delete_sales_orders(request, extended_mock_context)

    # Verify
    assert response.success is True
    assert "prod-123" in response.message
    mock_service.delete_for_product.assert_called_once_with("prod-123")


@pytest.mark.asyncio
async def test_delete_sales_orders_no_filter_raises_error(extended_mock_context):
    """Test deleting sales orders without filter raises error for safety."""
    # Execute and verify
    request = DeleteSalesOrdersRequest()
    with pytest.raises(ValueError, match="product_id is required for deletion"):
        await delete_sales_orders(request, extended_mock_context)



"""Purchase Order management tools for StockTrim MCP Server."""

from __future__ import annotations

import logging
from datetime import datetime

from fastmcp import Context, FastMCP
from pydantic import BaseModel, Field

from stocktrim_mcp_server.dependencies import get_services
from stocktrim_public_api_client.client_types import UNSET

logger = logging.getLogger(__name__)

# ============================================================================
# Tool 1: get_purchase_order
# ============================================================================


class GetPurchaseOrderRequest(BaseModel):
    """Request model for getting a purchase order."""

    reference_number: str = Field(..., description="Purchase order reference number")


class PurchaseOrderInfo(BaseModel):
    """Purchase order information."""

    reference_number: str
    supplier_code: str | None
    supplier_name: str | None
    status: str | None
    total_cost: float | None
    line_items_count: int


async def _get_purchase_order_impl(
    request: GetPurchaseOrderRequest, context: Context
) -> PurchaseOrderInfo | None:
    """Implementation of get_purchase_order tool.

    Args:
        request: Request containing reference number
        context: Server context with services

    Returns:
        PurchaseOrderInfo if found, None otherwise

    Raises:
        ValueError: If reference number is empty
        Exception: If API call fails
    """
    services = get_services(context)
    po = await services.purchase_orders.get_by_reference(request.reference_number)

    if not po:
        return None

    # Build PurchaseOrderInfo from response
    # Calculate total cost from line items
    total_cost = None
    if po.purchase_order_line_items:
        total_cost = sum(
            (item.unit_price or 0.0) * item.quantity
            for item in po.purchase_order_line_items
        )

    return PurchaseOrderInfo(
        reference_number=po.reference_number or "",
        supplier_code=po.supplier.supplier_code if po.supplier else None,
        supplier_name=po.supplier.supplier_name if po.supplier else None,
        status=str(po.status) if po.status else None,
        total_cost=total_cost,
        line_items_count=(
            len(po.purchase_order_line_items) if po.purchase_order_line_items else 0
        ),
    )


async def get_purchase_order(
    request: GetPurchaseOrderRequest, context: Context
) -> PurchaseOrderInfo | None:
    """Get a purchase order by reference number.

    This tool retrieves detailed information about a specific purchase order
    from StockTrim.

    Args:
        request: Request containing reference number
        context: Server context with StockTrimClient

    Returns:
        PurchaseOrderInfo if found, None if not found

    Example:
        Request: {"reference_number": "PO-2024-001"}
        Returns: {"reference_number": "PO-2024-001", "supplier_code": "SUP-001", ...}
    """
    return await _get_purchase_order_impl(request, context)


# ============================================================================
# Tool 2: list_purchase_orders
# ============================================================================


class ListPurchaseOrdersRequest(BaseModel):
    """Request model for listing purchase orders."""

    pass  # No filters for now, V1 API doesn't support filtering


class ListPurchaseOrdersResponse(BaseModel):
    """Response containing purchase orders."""

    purchase_orders: list[PurchaseOrderInfo]
    total_count: int


async def _list_purchase_orders_impl(
    request: ListPurchaseOrdersRequest, context: Context
) -> ListPurchaseOrdersResponse:
    """Implementation of list_purchase_orders tool.

    Args:
        request: Request (no filters supported yet)
        context: Server context with services

    Returns:
        ListPurchaseOrdersResponse with purchase orders

    Raises:
        Exception: If API call fails
    """
    services = get_services(context)
    pos = await services.purchase_orders.list_all()

    # Handle case where API returns single object instead of list
    if not isinstance(pos, list):
        pos = [pos] if pos else []

    # Build response
    po_infos = []
    for po in pos:
        # Calculate total cost from line items
        total_cost = None
        if po.purchase_order_line_items:
            total_cost = sum(
                (item.unit_price or 0.0) * item.quantity
                for item in po.purchase_order_line_items
            )

        po_infos.append(
            PurchaseOrderInfo(
                reference_number=po.reference_number or "",
                supplier_code=po.supplier.supplier_code if po.supplier else None,
                supplier_name=po.supplier.supplier_name if po.supplier else None,
                status=str(po.status) if po.status else None,
                total_cost=total_cost,
                line_items_count=(
                    len(po.purchase_order_line_items)
                    if po.purchase_order_line_items
                    else 0
                ),
            )
        )

    return ListPurchaseOrdersResponse(
        purchase_orders=po_infos,
        total_count=len(po_infos),
    )


async def list_purchase_orders(
    request: ListPurchaseOrdersRequest, context: Context
) -> ListPurchaseOrdersResponse:
    """List all purchase orders.

    This tool retrieves all purchase orders from StockTrim (V1 API).

    Args:
        request: Request (no filters supported yet)
        context: Server context with StockTrimClient

    Returns:
        ListPurchaseOrdersResponse with purchase orders

    Example:
        Request: {}
        Returns: {"purchase_orders": [...], "total_count": 15}
    """
    return await _list_purchase_orders_impl(request, context)


# ============================================================================
# Tool 3: create_purchase_order
# ============================================================================


class LineItemRequest(BaseModel):
    """Line item for purchase order."""

    product_code: str = Field(..., description="Product code")
    quantity: float = Field(..., description="Quantity to order", gt=0)
    unit_price: float | None = Field(default=None, description="Unit price")


class CreatePurchaseOrderRequest(BaseModel):
    """Request model for creating a purchase order."""

    supplier_code: str = Field(..., description="Supplier code")
    supplier_name: str | None = Field(default=None, description="Supplier name")
    line_items: list[LineItemRequest] = Field(
        ..., description="Line items for the purchase order", min_length=1
    )
    order_date: datetime | None = Field(
        default=None,
        description="Order date (ISO format). Defaults to current date if not provided.",
    )
    location_code: str | None = Field(default=None, description="Location code")
    location_name: str | None = Field(default=None, description="Location name")
    reference_number: str | None = Field(
        default=None, description="Custom reference number"
    )
    client_reference_number: str | None = Field(
        default=None, description="Client reference number"
    )
    status: str | None = Field(
        default="Draft",
        description="Purchase order status (Draft, Approved, Sent, Received)",
    )


class CreatePurchaseOrderResponse(BaseModel):
    """Response for purchase order creation."""

    reference_number: str
    supplier_code: str | None
    supplier_name: str | None
    status: str | None
    total_cost: float | None
    line_items_count: int


async def _create_purchase_order_impl(
    request: CreatePurchaseOrderRequest, context: Context
) -> CreatePurchaseOrderResponse:
    """Implementation of create_purchase_order tool.

    Args:
        request: Request containing purchase order details
        context: Server context with services

    Returns:
        CreatePurchaseOrderResponse with created PO details

    Raises:
        Exception: If API call fails
    """
    services = get_services(context)

    # Convert line items from pydantic models to dicts for service layer
    line_items = [
        {
            "product_code": item.product_code,
            "quantity": item.quantity,
            "unit_price": item.unit_price,
        }
        for item in request.line_items
    ]

    # Create purchase order via service
    created_po = await services.purchase_orders.create(
        supplier_code=request.supplier_code,
        line_items=line_items,
        supplier_name=request.supplier_name,
        order_date=request.order_date,
        location_code=request.location_code,
        location_name=request.location_name,
        reference_number=request.reference_number,
        client_reference_number=request.client_reference_number,
        status=request.status,
    )

    # Build response
    # Calculate total cost from line items
    total_cost = None
    if created_po.purchase_order_line_items:
        total_cost = sum(
            (item.unit_price or 0.0) * item.quantity
            for item in created_po.purchase_order_line_items
        )

    return CreatePurchaseOrderResponse(
        reference_number=created_po.reference_number or "",
        supplier_code=(
            created_po.supplier.supplier_code if created_po.supplier else None
        ),
        supplier_name=(
            created_po.supplier.supplier_name if created_po.supplier else None
        ),
        status=str(created_po.status) if created_po.status else None,
        total_cost=total_cost,
        line_items_count=(
            len(created_po.purchase_order_line_items)
            if created_po.purchase_order_line_items
            else 0
        ),
    )


async def create_purchase_order(
    request: CreatePurchaseOrderRequest, context: Context
) -> CreatePurchaseOrderResponse:
    """Create a new purchase order.

    This tool creates a new purchase order in StockTrim.

    Args:
        request: Request containing purchase order details
        context: Server context with StockTrimClient

    Returns:
        CreatePurchaseOrderResponse with created PO details

    Example:
        Request: {
            "supplier_code": "SUP-001",
            "supplier_name": "Acme Supplies",
            "line_items": [
                {"product_code": "WIDGET-001", "quantity": 100, "unit_price": 15.50}
            ],
            "status": "Draft"
        }
        Returns: {
            "reference_number": "PO-2024-001",
            "supplier_code": "SUP-001",
            "status": "Draft",
            "total_cost": 1550.0,
            "line_items_count": 1,
            "message": "Purchase order PO-2024-001 created successfully"
        }
    """
    return await _create_purchase_order_impl(request, context)


# ============================================================================
# Tool 4: delete_purchase_order
# ============================================================================


class DeletePurchaseOrderRequest(BaseModel):
    """Request model for deleting a purchase order."""

    reference_number: str = Field(..., description="Reference number to delete")


class DeletePurchaseOrderResponse(BaseModel):
    """Response for purchase order deletion."""

    success: bool
    message: str


async def _delete_purchase_order_impl(
    request: DeletePurchaseOrderRequest, context: Context
) -> DeletePurchaseOrderResponse:
    """Implementation of delete_purchase_order tool.

    Args:
        request: Request containing reference number
        context: Server context with services

    Returns:
        DeletePurchaseOrderResponse indicating success

    Raises:
        ValueError: If reference number is empty
        Exception: If API call fails
    """
    services = get_services(context)
    success, message = await services.purchase_orders.delete(request.reference_number)

    return DeletePurchaseOrderResponse(
        success=success,
        message=message,
    )


async def delete_purchase_order(
    request: DeletePurchaseOrderRequest, context: Context
) -> DeletePurchaseOrderResponse:
    """Delete a purchase order by reference number.

    This tool deletes a purchase order from StockTrim.

    Args:
        request: Request containing reference number
        context: Server context with StockTrimClient

    Returns:
        DeletePurchaseOrderResponse indicating success

    Example:
        Request: {"reference_number": "PO-2024-001"}
        Returns: {"success": true, "message": "Purchase order PO-2024-001 deleted successfully"}
    """
    return await _delete_purchase_order_impl(request, context)


# ============================================================================
# Tool Registration
# ============================================================================


def register_tools(mcp: FastMCP) -> None:
    """Register purchase order tools with FastMCP server.

    Args:
        mcp: FastMCP server instance
    """
    mcp.tool()(get_purchase_order)
    mcp.tool()(list_purchase_orders)
    mcp.tool()(create_purchase_order)
    mcp.tool()(delete_purchase_order)



from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, cast

from attrs import define as _attrs_define

from ...client_types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.products_response_dto import ProductsResponseDto


T = TypeVar("T", bound="BillOfMaterialsResponseDto")


@_attrs_define
class BillOfMaterialsResponseDto:
    """
    Attributes:
        product_id (str):
        component_id (str):
        id (int | Unset):
        assembly_time_days (int | None | Unset):
        sku_component (ProductsResponseDto | Unset):
        sku_product (ProductsResponseDto | Unset):
        quantity (float | None | Unset):
    """

    product_id: str
    component_id: str
    id: int | Unset = UNSET
    assembly_time_days: int | None | Unset = UNSET
    sku_component: ProductsResponseDto | Unset = UNSET
    sku_product: ProductsResponseDto | Unset = UNSET
    quantity: float | None | Unset = UNSET

    def to_dict(self) -> dict[str, Any]:
        product_id = self.product_id

        component_id = self.component_id

        id = self.id

        assembly_time_days: int | None | Unset
        if isinstance(self.assembly_time_days, Unset):
            assembly_time_days = UNSET
        else:
            assembly_time_days = self.assembly_time_days

        sku_component: dict[str, Any] | Unset = UNSET
        if not isinstance(self.sku_component, Unset):
            sku_component = self.sku_component.to_dict()

        sku_product: dict[str, Any] | Unset = UNSET
        if not isinstance(self.sku_product, Unset):
            sku_product = self.sku_product.to_dict()

        quantity: float | None | Unset
        if isinstance(self.quantity, Unset):
            quantity = UNSET
        else:
            quantity = self.quantity

        field_dict: dict[str, Any] = {}

        field_dict.update(
            {
                "productId": product_id,
                "componentId": component_id,
            }
        )
        if id is not UNSET:
            field_dict["id"] = id
        if assembly_time_days is not UNSET:
            field_dict["assemblyTimeDays"] = assembly_time_days
        if sku_component is not UNSET:
            field_dict["skuComponent"] = sku_component
        if sku_product is not UNSET:
            field_dict["skuProduct"] = sku_product
        if quantity is not UNSET:
            field_dict["quantity"] = quantity

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.products_response_dto import ProductsResponseDto

        d = dict(src_dict)
        product_id = d.pop("productId")

        component_id = d.pop("componentId")

        id = d.pop("id", UNSET)

        def _parse_assembly_time_days(data: object) -> int | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(int | None | Unset, data)

        assembly_time_days = _parse_assembly_time_days(d.pop("assemblyTimeDays", UNSET))

        _sku_component = d.pop("skuComponent", UNSET)
        sku_component: ProductsResponseDto | Unset
        if isinstance(_sku_component, Unset):
            sku_component = UNSET
        else:
            sku_component = ProductsResponseDto.from_dict(_sku_component)

        _sku_product = d.pop("skuProduct", UNSET)
        sku_product: ProductsResponseDto | Unset
        if isinstance(_sku_product, Unset):
            sku_product = UNSET
        else:
            sku_product = ProductsResponseDto.from_dict(_sku_product)

        def _parse_quantity(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        quantity = _parse_quantity(d.pop("quantity", UNSET))

        bill_of_materials_response_dto = cls(
            product_id=product_id,
            component_id=component_id,
            id=id,
            assembly_time_days=assembly_time_days,
            sku_component=sku_component,
            sku_product=sku_product,
            quantity=quantity,
        )

        return bill_of_materials_response_dto



from __future__ import annotations

from collections.abc import Mapping
from typing import Any, TypeVar, cast

from attrs import define as _attrs_define

from ...client_types import UNSET, Unset

T = TypeVar("T", bound="PurchaseOrderLocation")


@_attrs_define
class PurchaseOrderLocation:
    """
    Attributes:
        location_code (None | str | Unset):
        location_name (None | str | Unset):
    """

    location_code: None | str | Unset = UNSET
    location_name: None | str | Unset = UNSET

    def to_dict(self) -> dict[str, Any]:
        location_code: None | str | Unset
        if isinstance(self.location_code, Unset):
            location_code = UNSET
        else:
            location_code = self.location_code

        location_name: None | str | Unset
        if isinstance(self.location_name, Unset):
            location_name = UNSET
        else:
            location_name = self.location_name

        field_dict: dict[str, Any] = {}

        field_dict.update({})
        if location_code is not UNSET:
            field_dict["locationCode"] = location_code
        if location_name is not UNSET:
            field_dict["locationName"] = location_name

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)

        def _parse_location_code(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        location_code = _parse_location_code(d.pop("locationCode", UNSET))

        def _parse_location_name(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        location_name = _parse_location_name(d.pop("locationName", UNSET))

        purchase_order_location = cls(
            location_code=location_code,
            location_name=location_name,
        )

        return purchase_order_location



#!/usr/bin/env python3
"""
Test script to check actual HTTP status code returned by DELETE /api/PurchaseOrders.

This script:
1. Creates a test purchase order
2. Deletes it via the API
3. Reports the actual HTTP status code returned

This helps us determine if the API has been updated to return 204 (as Joel mentioned)
or still returns 200 (as the OpenAPI spec currently documents).
"""

import asyncio
import logging
import sys
from datetime import datetime

from stocktrim_public_api_client import StockTrimClient
from stocktrim_public_api_client.generated.models import (
    ProductsRequestDto,
    PurchaseOrderLineItem,
    PurchaseOrderRequestDto,
    PurchaseOrderSupplier,
)

# Configure logging to see HTTP details
logging.basicConfig(
    level=logging.INFO,
    format="%(levelname)s: %(message)s",
)
logger = logging.getLogger(__name__)

# Enable DEBUG for stocktrim_client to see HTTP status codes
logging.getLogger("stocktrim_client").setLevel(logging.DEBUG)


async def test_delete_status_code():
    """Test the actual DELETE /api/PurchaseOrders status code."""
    # Initialize client (uses environment variables for auth)
    client = StockTrimClient()

    try:
        # Generate unique reference for test
        test_ref = f"TEST-DELETE-{datetime.now().strftime('%Y%m%d-%H%M%S')}"

        logger.info(f"\n{'=' * 60}")
        logger.info("STEP 1: Create test product")
        logger.info(f"{'=' * 60}")

        # Create test product first
        test_product_id = f"TEST-DEL-{datetime.now().strftime('%Y%m%d%H%M%S')}"
        product_request = ProductsRequestDto(
            product_id=test_product_id,
            product_code_readable=test_product_id,
            name="Test Product for DELETE",
        )

        try:
            created_product = await client.products.create(product_request)
            logger.info(f"✅ Created product: {created_product.product_id}")
        except Exception as e:
            # Product might already exist - that's okay
            logger.info(f"Product may already exist: {e}")

        logger.info(f"\n{'=' * 60}")
        logger.info("STEP 2: Create test purchase order")
        logger.info(f"{'=' * 60}")

        # Create test purchase order
        po_request = PurchaseOrderRequestDto(
            client_reference_number=test_ref,
            order_date=datetime.now(),
            supplier=PurchaseOrderSupplier(
                supplier_name="Test Supplier for DELETE",
            ),
            purchase_order_line_items=[
                PurchaseOrderLineItem(
                    product_id=test_product_id,
                    quantity=1.0,
                )
            ],
        )

        created_po = await client.purchase_orders.create(po_request)
        logger.info(f"✅ Created PO: {created_po.client_reference_number}")

        logger.info(f"\n{'=' * 60}")
        logger.info("STEP 3: Delete purchase order and capture status code")
        logger.info(f"{'=' * 60}")

        # Delete the purchase order
        # The delete method doesn't expose the status code directly,
        # so we need to check the underlying httpx response
        delete_response = await client.purchase_orders.delete(test_ref)

        logger.info(f"\n{'=' * 60}")
        logger.info("RESULTS")
        logger.info(f"{'=' * 60}")

        # Check what we got back
        if delete_response is None:
            logger.info("✅ DELETE returned: None (typical for 204 No Content)")
            logger.info("\n📊 CONCLUSION: API likely returns 204 (spec needs updating)")
        else:
            logger.info(f"✅ DELETE returned: {delete_response}")
            logger.info(
                "\n📊 CONCLUSION: API returns 200 with content (matches current spec)"
            )

        logger.info("\n💡 Check DEBUG logs above for actual HTTP status code")

    except Exception as e:
        logger.error(f"\n❌ Test failed: {e}", exc_info=True)
        sys.exit(1)


if __name__ == "__main__":
    asyncio.run(test_delete_status_code())



"""MCP tools for StockTrim Inventory Management.

This module contains tool implementations that provide actions with side effects
for interacting with the StockTrim API.

Tool Organization:
------------------
- foundation/: Low-level, granular operations that map closely to API endpoints
- workflows/: High-level, intent-based tools that combine multiple operations (future)

Tool Registration Pattern:
--------------------------
Each tool module exports a register_tools(mcp) function that registers its tools
with the FastMCP instance. This avoids circular imports.

When adding new tool modules:
1. Create the new module in the appropriate directory
2. Define tools as regular async functions (no decorators)
3. Add a register_tools(mcp: FastMCP) function that calls mcp.tool() on each function
4. Import and call the registration function from this file
"""

from fastmcp import FastMCP

from .foundation import register_all_foundation_tools
from .workflows import register_all_workflow_tools


def register_all_tools(mcp: FastMCP) -> None:
    """Register all tools from all modules.

    Args:
        mcp: FastMCP server instance to register tools with
    """
    # Register foundation tools (low-level API operations)
    register_all_foundation_tools(mcp)

    # Register workflow tools (high-level intent-based operations)
    register_all_workflow_tools(mcp)


__all__ = [
    "register_all_tools",
]



"""Tests for SalesOrderService."""

from datetime import datetime
from unittest.mock import AsyncMock, MagicMock

import pytest

from stocktrim_mcp_server.services.sales_orders import SalesOrderService
from stocktrim_public_api_client.generated.models import SalesOrderResponseDto


@pytest.fixture
def mock_client():
    """Create a mock StockTrimClient."""
    client = MagicMock()
    client.sales_orders = MagicMock()
    client.sales_orders.create = AsyncMock()
    client.sales_orders.get_for_product = AsyncMock()
    client.sales_orders.get_all = AsyncMock()
    client.sales_orders.delete_for_product = AsyncMock()
    return client


@pytest.fixture
def sales_order_service(mock_client):
    """Create a SalesOrderService with mock client."""
    return SalesOrderService(mock_client)


@pytest.fixture
def sample_sales_order():
    """Create a sample sales order for testing."""
    return SalesOrderResponseDto(
        id=123,
        product_id="WIDGET-001",
        order_date=datetime(2025, 1, 15),
        quantity=10.0,
        external_reference_id="SO-001",
    )


# ============================================================================
# Test create
# ============================================================================


@pytest.mark.asyncio
async def test_create_success(sales_order_service, mock_client, sample_sales_order):
    """Test successfully creating a sales order."""
    mock_client.sales_orders.create.return_value = sample_sales_order

    result = await sales_order_service.create(
        product_id="WIDGET-001",
        order_date=datetime(2025, 1, 15),
        quantity=10.0,
    )

    assert result == sample_sales_order
    assert result.product_id == "WIDGET-001"
    assert result.quantity == 10.0
    mock_client.sales_orders.create.assert_called_once()


@pytest.mark.asyncio
async def test_create_with_all_params(sales_order_service, mock_client):
    """Test creating a sales order with all optional parameters."""
    created_order = SalesOrderResponseDto(
        id=123,
        product_id="WIDGET-001",
        order_date=datetime(2025, 1, 15),
        quantity=10.0,
        external_reference_id="SO-001",
        unit_price=20.0,
    )
    mock_client.sales_orders.create.return_value = created_order

    result = await sales_order_service.create(
        product_id="WIDGET-001",
        order_date=datetime(2025, 1, 15),
        quantity=10.0,
        external_reference_id="SO-001",
        unit_price=20.0,
        location_code="LOC-001",
        location_name="Warehouse 1",
        customer_code="CUST-001",
        customer_name="Test Customer",
    )

    assert result.product_id == "WIDGET-001"
    assert result.quantity == 10.0
    assert result.unit_price == 20.0
    assert result.external_reference_id == "SO-001"
    mock_client.sales_orders.create.assert_called_once()


@pytest.mark.asyncio
async def test_create_empty_product_id(sales_order_service):
    """Test error when product_id is empty."""
    with pytest.raises(ValueError, match="Product ID cannot be empty"):
        await sales_order_service.create(
            product_id="",
            order_date=datetime(2025, 1, 15),
            quantity=10.0,
        )

    with pytest.raises(ValueError, match="Product ID cannot be empty"):
        await sales_order_service.create(
            product_id="   ",
            order_date=datetime(2025, 1, 15),
            quantity=10.0,
        )


@pytest.mark.asyncio
async def test_create_invalid_quantity_zero(sales_order_service):
    """Test error when quantity is zero."""
    with pytest.raises(ValueError, match="Quantity must be greater than 0"):
        await sales_order_service.create(
            product_id="WIDGET-001",
            order_date=datetime(2025, 1, 15),
            quantity=0,
        )


@pytest.mark.asyncio
async def test_create_invalid_quantity_negative(sales_order_service):
    """Test error when quantity is negative."""
    with pytest.raises(ValueError, match="Quantity must be greater than 0"):
        await sales_order_service.create(
            product_id="WIDGET-001",
            order_date=datetime(2025, 1, 15),
            quantity=-5.0,
        )


@pytest.mark.asyncio
async def test_create_invalid_quantity_none(sales_order_service):
    """Test error when quantity is None."""
    with pytest.raises(ValueError, match="Quantity must be greater than 0"):
        await sales_order_service.create(
            product_id="WIDGET-001",
            order_date=datetime(2025, 1, 15),
            quantity=None,
        )


@pytest.mark.asyncio
async def test_create_api_error(sales_order_service, mock_client):
    """Test handling of API errors during creation."""
    mock_client.sales_orders.create.side_effect = Exception("API Error")

    with pytest.raises(Exception, match="API Error"):
        await sales_order_service.create(
            product_id="WIDGET-001",
            order_date=datetime(2025, 1, 15),
            quantity=10.0,
        )


# ============================================================================
# Test get_for_product
# ============================================================================


@pytest.mark.asyncio
async def test_get_for_product_success(sales_order_service, mock_client):
    """Test successfully getting sales orders for a product."""
    orders = [
        SalesOrderResponseDto(
            id=1,
            product_id="WIDGET-001",
            order_date=datetime(2025, 1, 15),
            quantity=10.0,
        ),
        SalesOrderResponseDto(
            id=2,
            product_id="WIDGET-001",
            order_date=datetime(2025, 1, 16),
            quantity=20.0,
        ),
    ]
    mock_client.sales_orders.get_for_product.return_value = orders

    result = await sales_order_service.get_for_product("WIDGET-001")

    assert len(result) == 2
    assert result[0].product_id == "WIDGET-001"
    assert result[1].product_id == "WIDGET-001"
    mock_client.sales_orders.get_for_product.assert_called_once_with("WIDGET-001")


@pytest.mark.asyncio
async def test_get_for_product_not_found(sales_order_service, mock_client):
    """Test getting sales orders for a product with no results."""
    mock_client.sales_orders.get_for_product.return_value = []

    result = await sales_order_service.get_for_product("NONEXISTENT")

    assert len(result) == 0
    mock_client.sales_orders.get_for_product.assert_called_once_with("NONEXISTENT")


@pytest.mark.asyncio
async def test_get_for_product_empty_product_id(sales_order_service):
    """Test error when product_id is empty."""
    with pytest.raises(ValueError, match="Product ID cannot be empty"):
        await sales_order_service.get_for_product("")

    with pytest.raises(ValueError, match="Product ID cannot be empty"):
        await sales_order_service.get_for_product("   ")


@pytest.mark.asyncio
async def test_get_for_product_api_error(sales_order_service, mock_client):
    """Test handling of API errors."""
    mock_client.sales_orders.get_for_product.side_effect = Exception("API Error")

    with pytest.raises(Exception, match="API Error"):
        await sales_order_service.get_for_product("WIDGET-001")


# ============================================================================
# Test get_all
# ============================================================================


@pytest.mark.asyncio
async def test_get_all_success(sales_order_service, mock_client):
    """Test successfully getting all sales orders."""
    orders = [
        SalesOrderResponseDto(
            id=1,
            product_id="WIDGET-001",
            order_date=datetime(2025, 1, 15),
            quantity=10.0,
        ),
        SalesOrderResponseDto(
            id=2,
            product_id="WIDGET-002",
            order_date=datetime(2025, 1, 16),
            quantity=20.0,
        ),
    ]
    mock_client.sales_orders.get_all.return_value = orders

    result = await sales_order_service.get_all()

    assert len(result) == 2
    assert result[0].product_id == "WIDGET-001"
    assert result[1].product_id == "WIDGET-002"
    mock_client.sales_orders.get_all.assert_called_once()


@pytest.mark.asyncio
async def test_get_all_empty(sales_order_service, mock_client):
    """Test getting all sales orders when none exist."""
    mock_client.sales_orders.get_all.return_value = []

    result = await sales_order_service.get_all()

    assert len(result) == 0
    mock_client.sales_orders.get_all.assert_called_once()


@pytest.mark.asyncio
async def test_get_all_filtered_by_product(sales_order_service, mock_client):
    """Test getting sales orders filtered by product."""
    orders = [
        SalesOrderResponseDto(
            id=1,
            product_id="WIDGET-001",
            order_date=datetime(2025, 1, 15),
            quantity=10.0,
        ),
        SalesOrderResponseDto(
            id=2,
            product_id="WIDGET-001",
            order_date=datetime(2025, 1, 16),
            quantity=20.0,
        ),
    ]
    mock_client.sales_orders.get_for_product.return_value = orders

    result = await sales_order_service.get_all(product_id="WIDGET-001")

    assert len(result) == 2
    assert result[0].product_id == "WIDGET-001"
    assert result[1].product_id == "WIDGET-001"
    # Should call get_for_product when product_id is provided
    mock_client.sales_orders.get_for_product.assert_called_once_with("WIDGET-001")
    mock_client.sales_orders.get_all.assert_not_called()


@pytest.mark.asyncio
async def test_get_all_api_error(sales_order_service, mock_client):
    """Test handling of API errors."""
    mock_client.sales_orders.get_all.side_effect = Exception("API Error")

    with pytest.raises(Exception, match="API Error"):
        await sales_order_service.get_all()


# ============================================================================
# Test delete_for_product
# ============================================================================


@pytest.mark.asyncio
async def test_delete_for_product_success(sales_order_service, mock_client):
    """Test successfully deleting sales orders for a product."""
    success, message = await sales_order_service.delete_for_product("WIDGET-001")

    assert success is True
    assert "deleted successfully" in message
    assert "WIDGET-001" in message
    mock_client.sales_orders.delete_for_product.assert_called_once_with("WIDGET-001")


@pytest.mark.asyncio
async def test_delete_for_product_empty_product_id(sales_order_service):
    """Test error when product_id is empty."""
    with pytest.raises(ValueError, match="Product ID cannot be empty"):
        await sales_order_service.delete_for_product("")

    with pytest.raises(ValueError, match="Product ID cannot be empty"):
        await sales_order_service.delete_for_product("   ")


@pytest.mark.asyncio
async def test_delete_for_product_api_error(sales_order_service, mock_client):
    """Test handling of API errors during deletion."""
    mock_client.sales_orders.delete_for_product.side_effect = Exception("API Error")

    with pytest.raises(Exception, match="API Error"):
        await sales_order_service.delete_for_product("WIDGET-001")



"""Customer management tools for StockTrim MCP Server."""

from __future__ import annotations

import logging

from fastmcp import Context, FastMCP
from pydantic import BaseModel, Field

from stocktrim_mcp_server.dependencies import get_services

logger = logging.getLogger(__name__)

# ============================================================================
# Tool 1: get_customer
# ============================================================================


class GetCustomerRequest(BaseModel):
    """Request model for getting a customer."""

    code: str = Field(..., description="Customer code to retrieve")


class CustomerInfo(BaseModel):
    """Customer information."""

    code: str
    name: str | None
    email: str | None
    phone: str | None
    address: str | None


async def get_customer(
    request: GetCustomerRequest, context: Context
) -> CustomerInfo | None:
    """Get a customer by code.

    This tool retrieves detailed information about a specific customer
    from StockTrim.

    Args:
        request: Request containing customer code
        context: Server context with StockTrimClient

    Returns:
        CustomerInfo if found, None if not found

    Example:
        Request: {"code": "CUST-001"}
        Returns: {"code": "CUST-001", "name": "Customer Name", ...}
    """
    services = get_services(context)
    customer = await services.customers.get_by_code(request.code)

    if not customer:
        return None

    # Build CustomerInfo from response
    return CustomerInfo(
        code=customer.code or "",
        name=customer.name,
        email=customer.email_address,
        phone=customer.phone,
        address=customer.street_address,
    )


# ============================================================================
# Tool 2: list_customers
# ============================================================================


class ListCustomersRequest(BaseModel):
    """Request model for listing customers."""

    limit: int = Field(default=50, description="Maximum customers to return")


class ListCustomersResponse(BaseModel):
    """Response containing customers."""

    customers: list[CustomerInfo]
    total_count: int


async def list_customers(
    request: ListCustomersRequest, context: Context
) -> ListCustomersResponse:
    """List all customers.

    This tool retrieves a list of all customers from StockTrim.
    Results are limited by the limit parameter.

    Args:
        request: Request with limit
        context: Server context with StockTrimClient

    Returns:
        ListCustomersResponse with customers

    Example:
        Request: {"limit": 50}
        Returns: {"customers": [...], "total_count": 50}
    """
    services = get_services(context)
    customers = await services.customers.list_all(limit=request.limit)

    # Build response
    customer_infos = [
        CustomerInfo(
            code=c.code or "",
            name=c.name,
            email=c.email_address,
            phone=c.phone,
            address=c.street_address,
        )
        for c in customers
    ]

    return ListCustomersResponse(
        customers=customer_infos,
        total_count=len(customer_infos),
    )


# ============================================================================
# Tool Registration
# ============================================================================


def register_tools(mcp: FastMCP) -> None:
    """Register customer tools with FastMCP server.

    Args:
        mcp: FastMCP server instance
    """
    mcp.tool()(get_customer)
    mcp.tool()(list_customers)



from http import HTTPStatus
from typing import Any, cast

import httpx

from ....client_types import Response
from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.problem_details import ProblemDetails
from ...models.products_request_dto import ProductsRequestDto
from ...models.products_response_dto import ProductsResponseDto


def _get_kwargs(
    *,
    body: ProductsRequestDto | ProductsRequestDto | ProductsRequestDto,
) -> dict[str, Any]:
    headers: dict[str, Any] = {}

    _kwargs: dict[str, Any] = {
        "method": "post",
        "url": "/api/Products",
    }

    if isinstance(body, ProductsRequestDto):
        _kwargs["json"] = body.to_dict()

        headers["Content-Type"] = "application/json-patch+json"
    if isinstance(body, ProductsRequestDto):
        _kwargs["json"] = body.to_dict()

        headers["Content-Type"] = "application/json"
    if isinstance(body, ProductsRequestDto):
        _kwargs["json"] = body.to_dict()

        headers["Content-Type"] = "application/*+json"

    _kwargs["headers"] = headers
    return _kwargs


def _parse_response(
    *, client: AuthenticatedClient | Client, response: httpx.Response
) -> Any | ProblemDetails | ProductsResponseDto | None:
    if response.status_code == 200:
        response_200 = ProductsResponseDto.from_dict(response.json())

        return response_200

    if response.status_code == 201:
        response_201 = ProductsResponseDto.from_dict(response.json())

        return response_201

    if response.status_code == 400:
        response_400 = ProblemDetails.from_dict(response.json())

        return response_400

    if response.status_code == 500:
        response_500 = cast(Any, None)
        return response_500

    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: AuthenticatedClient | Client, response: httpx.Response
) -> Response[Any | ProblemDetails | ProductsResponseDto]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    *,
    client: AuthenticatedClient | Client,
    body: ProductsRequestDto | ProductsRequestDto | ProductsRequestDto,
) -> Response[Any | ProblemDetails | ProductsResponseDto]:
    """
    Args:
        body (ProductsRequestDto):
        body (ProductsRequestDto):
        body (ProductsRequestDto):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Any | ProblemDetails | ProductsResponseDto]
    """

    kwargs = _get_kwargs(
        body=body,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    *,
    client: AuthenticatedClient | Client,
    body: ProductsRequestDto | ProductsRequestDto | ProductsRequestDto,
) -> Any | ProblemDetails | ProductsResponseDto | None:
    """
    Args:
        body (ProductsRequestDto):
        body (ProductsRequestDto):
        body (ProductsRequestDto):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Any | ProblemDetails | ProductsResponseDto
    """

    return sync_detailed(
        client=client,
        body=body,
    ).parsed


async def asyncio_detailed(
    *,
    client: AuthenticatedClient | Client,
    body: ProductsRequestDto | ProductsRequestDto | ProductsRequestDto,
) -> Response[Any | ProblemDetails | ProductsResponseDto]:
    """
    Args:
        body (ProductsRequestDto):
        body (ProductsRequestDto):
        body (ProductsRequestDto):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Any | ProblemDetails | ProductsResponseDto]
    """

    kwargs = _get_kwargs(
        body=body,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    *,
    client: AuthenticatedClient | Client,
    body: ProductsRequestDto | ProductsRequestDto | ProductsRequestDto,
) -> Any | ProblemDetails | ProductsResponseDto | None:
    """
    Args:
        body (ProductsRequestDto):
        body (ProductsRequestDto):
        body (ProductsRequestDto):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Any | ProblemDetails | ProductsResponseDto
    """

    return (
        await asyncio_detailed(
            client=client,
            body=body,
        )
    ).parsed



#!/usr/bin/env python3
"""Test DELETE /api/Products status code."""

import asyncio
import logging
from datetime import datetime

from stocktrim_public_api_client import StockTrimClient
from stocktrim_public_api_client.generated.models import ProductsRequestDto

logging.basicConfig(level=logging.INFO, format="%(levelname)s: %(message)s")
logger = logging.getLogger(__name__)
logging.getLogger("stocktrim_client").setLevel(logging.DEBUG)


async def main():
    """Test DELETE /api/Products status code."""
    client = StockTrimClient()

    try:
        # Create test product
        test_id = f"TEST-DEL-{datetime.now().strftime('%Y%m%d%H%M%S')}"
        logger.info(f"Creating test product: {test_id}")

        product = ProductsRequestDto(
            product_id=test_id,
            product_code_readable=test_id,
            name="Test Product for DELETE",
        )

        await client.products.create(product)
        logger.info(f"✅ Created product: {test_id}")

        # Delete it
        logger.info(f"Deleting product: {test_id}")
        logger.info("About to call client.products.delete()...")
        try:
            result = await client.products.delete(test_id)
            logger.info("✅ DELETE completed successfully")
            logger.info(f"Result type: {type(result)}")
            logger.info(f"Result value: {result}")
        except Exception as e:
            logger.error(f"❌ DELETE failed: {e}", exc_info=True)

    finally:
        pass


if __name__ == "__main__":
    asyncio.run(main())



"""Inventory management service."""

from __future__ import annotations

import logging

from stocktrim_mcp_server.services.base import BaseService
from stocktrim_public_api_client.client_types import UNSET, Unset
from stocktrim_public_api_client.generated.models import PurchaseOrderResponseDto

logger = logging.getLogger(__name__)


class InventoryService(BaseService):
    """Service for inventory management operations."""

    async def set_for_product(
        self,
        product_id: str,
        stock_on_hand: float | None = None,
        stock_on_order: float | None = None,
        location_code: str | None = None,
        location_name: str | None = None,
    ) -> PurchaseOrderResponseDto:
        """Set inventory levels for a product.

        Args:
            product_id: Product ID to set inventory for
            stock_on_hand: Current stock on hand quantity (optional)
            stock_on_order: Stock on order quantity (optional)
            location_code: Location code (optional)
            location_name: Location name (optional)

        Returns:
            PurchaseOrderResponseDto object (API inconsistency - API returns this
            type for inventory operations)

        Raises:
            ValueError: If product_id is empty
            Exception: If API call fails
        """
        self.validate_not_empty(product_id, "Product ID")
        logger.info(f"Setting inventory for product: {product_id}")

        # Convert None to UNSET for optional parameters
        stock_on_hand_param: float | Unset = (
            stock_on_hand if stock_on_hand is not None else UNSET
        )
        stock_on_order_param: float | Unset = (
            stock_on_order if stock_on_order is not None else UNSET
        )
        location_code_param: str | None | Unset = (
            location_code if location_code is not None else UNSET
        )
        location_name_param: str | None | Unset = (
            location_name if location_name is not None else UNSET
        )

        result = await self._client.inventory.set_for_product(
            product_id=product_id,
            stock_on_hand=stock_on_hand_param,
            stock_on_order=stock_on_order_param,
            location_code=location_code_param,
            location_name=location_name_param,
        )

        logger.info(f"Inventory set successfully for product: {product_id}")
        return result



"""Tests for the StockTrim client."""

import logging
import os
import re
from unittest.mock import AsyncMock, MagicMock, patch

import httpx
import pytest

from stocktrim_public_api_client import StockTrimClient
from stocktrim_public_api_client.stocktrim_client import ErrorLoggingTransport


class TestStockTrimClient:
    """Test the StockTrimClient class."""

    def test_client_initialization_with_credentials(self, mock_api_credentials):
        """Test client can be initialized with credentials."""
        client = StockTrimClient(**mock_api_credentials)

        # New architecture: client inherits from AuthenticatedClient
        assert isinstance(client, StockTrimClient)
        assert client.base_url == "https://api.test.stocktrim.example.com"
        assert client.max_retries == 5

    def test_client_initialization_from_env(self, mock_env_credentials):
        """Test client can be initialized from environment variables."""
        client = StockTrimClient()

        # Verify client was created successfully
        assert isinstance(client, StockTrimClient)
        # The mock_env_credentials fixture sets a custom base URL
        assert client.base_url == "https://api.env.stocktrim.example.com"

    def test_client_missing_credentials_raises_error(self):
        """Test client raises error when credentials are missing."""
        # Clear all environment variables including StockTrim credentials
        with (
            patch.dict(os.environ, {}, clear=True),
            patch.dict(
                os.environ,
                {"STOCKTRIM_API_AUTH_ID": "", "STOCKTRIM_API_AUTH_SIGNATURE": ""},
                clear=True,
            ),
            pytest.raises(ValueError, match="API credentials required"),
        ):
            StockTrimClient()

    def test_client_repr(self, stocktrim_client):
        """Test client string representation."""
        repr_str = repr(stocktrim_client)
        assert "StockTrimClient" in repr_str
        assert "base_url" in repr_str
        assert "max_retries" in repr_str

    @pytest.mark.asyncio
    async def test_client_context_manager(self, stocktrim_client):
        """Test client works as async context manager."""
        async with stocktrim_client as client:
            assert client is stocktrim_client
            # New architecture: client IS the authenticated client
            assert isinstance(client, StockTrimClient)

        # Client should be properly closed after context exit
        # The underlying httpx client is closed by the parent class


class TestErrorLoggingTransport:
    """Test the ErrorLoggingTransport logging functionality."""

    @pytest.fixture
    def mock_transport(self):
        """Create a mock AsyncHTTPTransport."""
        transport = AsyncMock(spec=httpx.AsyncHTTPTransport)
        return transport

    @pytest.fixture
    def mock_logger(self):
        """Create a mock logger for testing."""
        logger = MagicMock(spec=logging.Logger)
        logger.isEnabledFor = MagicMock(return_value=True)
        return logger

    @pytest.fixture
    def logging_transport(self, mock_transport, mock_logger):
        """Create ErrorLoggingTransport with mocked dependencies."""
        return ErrorLoggingTransport(
            wrapped_transport=mock_transport, logger=mock_logger
        )

    @pytest.fixture
    def mock_request(self):
        """Create a mock HTTP request."""
        request = MagicMock(spec=httpx.Request)
        request.method = "GET"
        request.url = "https://api.stocktrim.com/api/Products"
        request.headers = {
            "user-agent": "test-client",
            "api-auth-id": "secret-id",
            "api-auth-signature": "secret-signature",
        }
        return request

    def create_mock_response(
        self, status_code: int, json_data=None, text=None, request=None
    ):
        """Create a mock HTTP response."""
        response = MagicMock(spec=httpx.Response)
        response.status_code = status_code
        response.request = request or MagicMock(spec=httpx.Request)

        if json_data is not None:
            response.json = MagicMock(return_value=json_data)
            response.text = str(json_data)
        elif text is not None:
            response.json = MagicMock(side_effect=ValueError("Not JSON"))
            response.text = text
        else:
            response.json = MagicMock(return_value=None)
            response.text = "null"

        # Mock aread for streaming responses
        response.aread = AsyncMock()
        return response

    @pytest.mark.asyncio
    async def test_log_request_sanitizes_auth_headers(
        self, logging_transport, mock_request, mock_logger
    ):
        """Test that request logging excludes auth headers."""
        await logging_transport._log_request(mock_request)

        # Verify debug was called
        assert mock_logger.debug.called
        call_args = mock_logger.debug.call_args

        # Check that auth headers were sanitized
        extra_data = call_args.kwargs.get("extra", {})
        headers = extra_data.get("headers", {})
        assert "user-agent" in headers
        assert "api-auth-id" not in headers
        assert "api-auth-signature" not in headers

    @pytest.mark.asyncio
    async def test_log_success_response_info_level(
        self, logging_transport, mock_request, mock_logger
    ):
        """Test that successful responses are logged at INFO level."""
        response = self.create_mock_response(
            200, json_data={"id": 123}, request=mock_request
        )

        await logging_transport._log_success_response(response, mock_request, 100.5)

        # Verify INFO log was called with status and timing
        mock_logger.info.assert_called_once()
        info_message = mock_logger.info.call_args[0][0]
        assert "GET" in info_message
        assert "200" in info_message
        # Use regex to match timing - avoid fragile exact millisecond checks
        assert re.search(r"\d+ms", info_message)

    @pytest.mark.asyncio
    async def test_log_success_response_debug_body_list(
        self, logging_transport, mock_request, mock_logger
    ):
        """Test DEBUG logging shows list length for array responses."""
        response = self.create_mock_response(
            200, json_data=[{"id": 1}, {"id": 2}, {"id": 3}], request=mock_request
        )
        mock_logger.isEnabledFor.return_value = True

        await logging_transport._log_success_response(response, mock_request, 50.0)

        # Verify DEBUG log shows list length
        debug_calls = [call[0][0] for call in mock_logger.debug.call_args_list]
        assert any("list[3] items" in call for call in debug_calls)

    @pytest.mark.asyncio
    async def test_log_success_response_null_body(
        self, logging_transport, mock_request, mock_logger
    ):
        """Test that null responses are logged at DEBUG level without extra warnings."""
        response = self.create_mock_response(200, json_data=None, request=mock_request)
        mock_logger.isEnabledFor.return_value = True

        await logging_transport._log_success_response(response, mock_request, 75.0)

        # Verify INFO log was called for the successful response
        mock_logger.info.assert_called_once()

        # Verify DEBUG log shows null body
        debug_calls = [call[0][0] for call in mock_logger.debug.call_args_list]
        assert any("null" in call.lower() for call in debug_calls)

        # Verify no WARNING or ERROR level logs for null responses
        # (TypeErrors will be logged separately via log_parsing_error when they occur)
        mock_logger.warning.assert_not_called()
        mock_logger.error.assert_not_called()

    @pytest.mark.asyncio
    async def test_log_success_response_dict_excerpt(
        self, logging_transport, mock_request, mock_logger
    ):
        """Test DEBUG logging shows dict excerpt for object responses."""
        large_dict = {"key": "value" * 100}  # Large response
        response = self.create_mock_response(
            200, json_data=large_dict, request=mock_request
        )
        mock_logger.isEnabledFor.return_value = True

        await logging_transport._log_success_response(response, mock_request, 50.0)

        # Verify DEBUG log shows truncated body
        debug_calls = [call[0][0] for call in mock_logger.debug.call_args_list]
        assert any("Response body:" in call and "..." in call for call in debug_calls)

    @pytest.mark.asyncio
    async def test_log_client_error_with_json(
        self, logging_transport, mock_request, mock_logger
    ):
        """Test 4xx error logging with JSON response."""
        error_data = {"type": "about:blank", "title": "Not Found", "status": 404}
        response = self.create_mock_response(
            404, json_data=error_data, request=mock_request
        )

        await logging_transport._log_client_error(response, mock_request, 125.5)

        # Verify ERROR was logged with error details and timing
        mock_logger.error.assert_called()
        error_message = mock_logger.error.call_args[0][0]
        assert "404" in error_message
        # Use regex to match timing - avoid fragile exact millisecond checks
        assert re.search(r"\d+ms", error_message)

    @pytest.mark.asyncio
    async def test_log_client_error_non_json(
        self, logging_transport, mock_request, mock_logger
    ):
        """Test 4xx error logging with non-JSON response."""
        response = self.create_mock_response(
            400, text="Bad Request", request=mock_request
        )

        await logging_transport._log_client_error(response, mock_request, 50.0)

        # Verify ERROR was logged with text response
        mock_logger.error.assert_called()
        error_message = mock_logger.error.call_args[0][0]
        assert "400" in error_message
        # Use regex to match timing
        assert re.search(r"\d+ms", error_message)

    @pytest.mark.asyncio
    async def test_log_server_error_with_json(
        self, logging_transport, mock_request, mock_logger
    ):
        """Test 5xx error logging with JSON response."""
        error_data = {"error": "Internal Server Error"}
        response = self.create_mock_response(
            500, json_data=error_data, request=mock_request
        )

        await logging_transport._log_server_error(response, mock_request, 200.0)

        # Verify ERROR was logged with server error details
        mock_logger.error.assert_called()
        error_message = mock_logger.error.call_args[0][0]
        assert "500" in error_message
        assert "Server error" in error_message
        # Use regex to match timing
        assert re.search(r"\d+ms", error_message)

    @pytest.mark.asyncio
    async def test_log_server_error_non_json(
        self, logging_transport, mock_request, mock_logger
    ):
        """Test 5xx error logging with non-JSON response."""
        response = self.create_mock_response(
            503, text="Service Unavailable", request=mock_request
        )

        await logging_transport._log_server_error(response, mock_request, 300.0)

        # Verify ERROR was logged with text response
        mock_logger.error.assert_called()
        error_message = mock_logger.error.call_args[0][0]
        assert "503" in error_message
        # Use regex to match timing
        assert re.search(r"\d+ms", error_message)

    @pytest.mark.asyncio
    async def test_handle_async_request_routes_2xx(
        self, logging_transport, mock_request, mock_transport, mock_logger
    ):
        """Test that 2xx responses are routed to success logging."""
        response = self.create_mock_response(
            200, json_data={"result": "ok"}, request=mock_request
        )
        mock_transport.handle_async_request.return_value = response

        result = await logging_transport.handle_async_request(mock_request)

        assert result == response
        mock_logger.info.assert_called()  # Should log at INFO level

    @pytest.mark.asyncio
    async def test_handle_async_request_routes_4xx(
        self, logging_transport, mock_request, mock_transport, mock_logger
    ):
        """Test that 4xx responses are routed to client error logging."""
        response = self.create_mock_response(
            404, json_data={"error": "Not found"}, request=mock_request
        )
        mock_transport.handle_async_request.return_value = response

        result = await logging_transport.handle_async_request(mock_request)

        assert result == response
        mock_logger.error.assert_called()  # Should log at ERROR level

    @pytest.mark.asyncio
    async def test_handle_async_request_routes_5xx(
        self, logging_transport, mock_request, mock_transport, mock_logger
    ):
        """Test that 5xx responses are routed to server error logging."""
        response = self.create_mock_response(
            500, json_data={"error": "Server error"}, request=mock_request
        )
        mock_transport.handle_async_request.return_value = response

        result = await logging_transport.handle_async_request(mock_request)

        assert result == response
        mock_logger.error.assert_called()  # Should log at ERROR level

    @pytest.mark.asyncio
    async def test_logging_respects_debug_level(
        self, logging_transport, mock_request, mock_logger
    ):
        """Test that DEBUG-level logging is skipped when DEBUG is disabled."""
        mock_logger.isEnabledFor.return_value = False
        response = self.create_mock_response(
            200, json_data={"data": "test"}, request=mock_request
        )

        await logging_transport._log_success_response(response, mock_request, 50.0)

        # INFO should still be called
        mock_logger.info.assert_called()
        # But DEBUG details should not be logged
        assert mock_logger.debug.call_count == 0

    def test_log_parsing_error_with_null_fields(
        self, logging_transport, mock_request, mock_logger
    ):
        """Test that log_parsing_error identifies null fields and provides fix suggestions."""
        response_data = {
            "id": 123,
            "orderDate": None,
            "fullyReceivedDate": None,
            "supplier": {"supplierName": None, "supplierCode": "SUP001"},
        }
        response = self.create_mock_response(200, json_data=response_data)

        error = TypeError("object of type 'NoneType' has no len()")
        logging_transport.log_parsing_error(error, response, mock_request)

        # Get all error messages
        error_messages = [call[0][0] for call in mock_logger.error.call_args_list]

        # Verify TypeError was identified
        assert any("TypeError" in msg for msg in error_messages)

        # Verify null fields were found
        assert any("3 null field(s)" in msg for msg in error_messages)
        assert any("orderDate" in msg for msg in error_messages)
        assert any("fullyReceivedDate" in msg for msg in error_messages)
        assert any("supplier.supplierName" in msg for msg in error_messages)

        # Verify fix suggestions are provided
        assert any("Possible fixes:" in msg for msg in error_messages)
        assert any("NULLABLE_FIELDS" in msg for msg in error_messages)
        assert any("regenerate_client.py" in msg for msg in error_messages)
        assert any("nullable: true" in msg for msg in error_messages)

        # Verify documentation link is provided
        assert any("api-feedback.md" in msg for msg in error_messages)

    def test_log_parsing_error_value_error(
        self, logging_transport, mock_request, mock_logger
    ):
        """Test that log_parsing_error shows response excerpt for ValueErrors."""
        response_data = {"id": "abc", "name": "Product 1"}
        response = self.create_mock_response(200, json_data=response_data)

        error = ValueError("invalid literal for int() with base 10: 'abc'")
        logging_transport.log_parsing_error(error, response, mock_request)

        # Verify error was logged
        assert mock_logger.error.call_count >= 2
        error_messages = [call[0][0] for call in mock_logger.error.call_args_list]

        # Verify ValueError was identified
        assert any("ValueError" in msg for msg in error_messages)

        # Verify response excerpt was shown
        assert any("Response excerpt:" in msg for msg in error_messages)

    def test_log_parsing_error_non_json_response(
        self, logging_transport, mock_request, mock_logger
    ):
        """Test that log_parsing_error handles non-JSON responses."""
        response = self.create_mock_response(200, text="Not JSON content")

        error = TypeError("expected dict, got str")
        logging_transport.log_parsing_error(error, response, mock_request)

        # Verify error was logged with response text
        assert mock_logger.error.call_count >= 3
        error_messages = [call[0][0] for call in mock_logger.error.call_args_list]

        # Verify error type was logged
        assert any("TypeError" in msg for msg in error_messages)

        # Verify response text was shown
        assert any("Response text:" in msg for msg in error_messages)



"""Dependency injection helpers for MCP tools."""

from __future__ import annotations

from fastmcp import Context

from stocktrim_mcp_server.context import ServerContext


def get_services(context: Context) -> ServerContext:
    """Extract ServerContext from FastMCP context.

    This helper function provides a clean way to access the service layer
    from MCP tool functions.

    Args:
        context: FastMCP context from tool invocation

    Returns:
        ServerContext with initialized services

    Example:
        async def get_product(request: GetProductRequest, context: Context) -> ProductInfo:
            services = get_services(context)
            return await services.products.get_by_code(request.code)
    """
    return context.request_context.lifespan_context



from __future__ import annotations

from collections.abc import Mapping
from typing import Any, TypeVar, cast

from attrs import define as _attrs_define

from ...client_types import UNSET, Unset

T = TypeVar("T", bound="ProductLocation")


@_attrs_define
class ProductLocation:
    """
    Attributes:
        location_code (str):
        location_name (None | str | Unset):
        stock_on_hand (float | None | Unset):
        stock_on_order (float | None | Unset):
    """

    location_code: str
    location_name: None | str | Unset = UNSET
    stock_on_hand: float | None | Unset = UNSET
    stock_on_order: float | None | Unset = UNSET

    def to_dict(self) -> dict[str, Any]:
        location_code = self.location_code

        location_name: None | str | Unset
        if isinstance(self.location_name, Unset):
            location_name = UNSET
        else:
            location_name = self.location_name

        stock_on_hand: float | None | Unset
        if isinstance(self.stock_on_hand, Unset):
            stock_on_hand = UNSET
        else:
            stock_on_hand = self.stock_on_hand

        stock_on_order: float | None | Unset
        if isinstance(self.stock_on_order, Unset):
            stock_on_order = UNSET
        else:
            stock_on_order = self.stock_on_order

        field_dict: dict[str, Any] = {}

        field_dict.update(
            {
                "locationCode": location_code,
            }
        )
        if location_name is not UNSET:
            field_dict["locationName"] = location_name
        if stock_on_hand is not UNSET:
            field_dict["stockOnHand"] = stock_on_hand
        if stock_on_order is not UNSET:
            field_dict["stockOnOrder"] = stock_on_order

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        location_code = d.pop("locationCode")

        def _parse_location_name(data: object) -> None | str | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(None | str | Unset, data)

        location_name = _parse_location_name(d.pop("locationName", UNSET))

        def _parse_stock_on_hand(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        stock_on_hand = _parse_stock_on_hand(d.pop("stockOnHand", UNSET))

        def _parse_stock_on_order(data: object) -> float | None | Unset:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(float | None | Unset, data)

        stock_on_order = _parse_stock_on_order(d.pop("stockOnOrder", UNSET))

        product_location = cls(
            location_code=location_code,
            location_name=location_name,
            stock_on_hand=stock_on_hand,
            stock_on_order=stock_on_order,
        )

        return product_location



"""Service layer for MCP tools."""

from stocktrim_mcp_server.services.base import BaseService
from stocktrim_mcp_server.services.inventory import InventoryService
from stocktrim_mcp_server.services.locations import LocationService
from stocktrim_mcp_server.services.products import ProductService
from stocktrim_mcp_server.services.purchase_orders import PurchaseOrderService
from stocktrim_mcp_server.services.sales_orders import SalesOrderService

__all__ = [
    "BaseService",
    "InventoryService",
    "LocationService",
    "ProductService",
    "PurchaseOrderService",
    "SalesOrderService",
]
